##
 # Makefile
 #
 # Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 # Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 #
 # out_yasapi-nt is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 #
 # out_yasapi-nt is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 ##
###############################################################################
include config.mak
include $(srcdir)/sources.mak
include $(srcdir)/version.mak

.PHONY: all install uninstall release FORCE

###############################################################################
VPATH=$(srcdir)

ifneq (ms,$(tools)) # [
CC=gcc
CFLAGS+=-O2
CFLAGS+=-shared
LDFLAGS+=-mwindows
#LDFLAGS+=-L$(PWD) # Not needed because there's a dependency to the delayload
                   # import libraries and hence they will be linked without
									 # searching.
EXTO=.o
EXTA=.a
EXTDLLA=.dll.a
CFLAGS+=-Wall
CFLAGS+=-Werror
LDLIBS+=-lavrt
LDLIBS+=-lksguid
LDLIBS+=-lole32
else # ] [
PATH+=:$(msvc)/bin
CC=cl
LIB=lib -nologo
EXTO=.obj
EXTA=.lib
EXTDLLA=.lib
CFLAGS+=-nologo
CFLAGS+=-O2
#CFLAGS+=-MT	# Causes the application to use the multithread, static
					# version of the run-time library. Defines _MT and causes
					# the compiler to place the library name LIBCMT.lib into
					# the .obj file so that the linker will use LIBCMT.lib to
					# resolve external symbols.
CFLAGS+=-MD	# Causes the application to use the multithread-specific and
					# DLL-specific version of the run-time library. Defines _MT
					# and _DLL and causes the compiler to place the library name
					# MSVCRT.lib into the .obj file.
CPPFLAGS+=-D_USING_V110_SDK71_
CPPFLAGS+='-I$(mssdk)/include'
CPPFLAGS+='-I$(msvc)/include'
LDFLAGS+='-libpath:$(mssdk)/lib'
LDFLAGS+='-libpath:$(msvc)/lib'
LDFLAGS+='-delayload:libgtk-3-0.dll'
#LDFLAGS+='-delayload:libgdk-3-0.dll'
#LDFLAGS+='-delayload:libgobject-2.0-0.dll'
#LDFLAGS+='-delayload:libgio-2.0-0.dll'
LDFLAGS+='-delayload:libglib-2.0-0.dll'
LDLIBS+=delayimp.lib
LDLIBS+=kernel32.lib
LDLIBS+=user32.lib
LDLIBS+=avrt.lib
LDLIBS+=ole32.lib
LDLIBS+=ksguid.lib
endif # ]

ifneq (,$(APPVER)) # [
CPPFLAGS+=-DYASAPI_VERSION=$(APPVER)
else # ] [
CPPFLAGS+=-DYASAPI_VERSION=0.0.0
endif # ]
CPPFLAGS+='-I$(srcdir)'
CPPFLAGS+='-I$(wasdk)'

GTK=gtk+-3.0
ifneq (,$(GTK)) # [
CPPFLAGS+=$(shell pkg-config --cflags-only-I $(GTK))
endif # ]

ifeq (ms,$(tools)) # [
yasapi_uuid$(EXTO): yasapi_uuid.c
	$(CC) $(CFLAGS) $(CPPFLAGS) -Tp $< -c -Fo$@
else # ] [
###############################################################################
ifneq (,$(GTK)) # [
CFLAGS+=$(shell pkg-config --cflags-only-other $(GTK))
#CPPFLAGS+=$(shell pkg-config --cflags-only-I $(GTK))
#LDFLAGS+=$(shell pkg-config --libs-only-L $(GTK))
#LDLIBS+=$(shell pkg-config --libs-only-l $(GTK))
endif # ]
endif # ]

###############################################################################
plugins=$(winamp)/Plugins
ALL+=out_yasapi.dll
#ifneq (ms,$(tools)) # [
ALL+=x7
x7: x7$(EXTO)
#ALL+=x6
ALL+=x5
ALL+=x4
ALL+=x3
ALL+=x2
ALL+=x1
#endif # ]
$(info $(ALL))
all: $(ALL)
#ifneq (ms,$(tools)) # [
#install: gtk-installed.stamp
#endif # ]
install: installed.stamp
installed.stamp: FORCE
installed.stamp: out_yasapi.dll
	mkdir -p '$(plugins)'
	cp -p $< '$(plugins)/$(app).dll'
	touch $@
gtk-installed.stamp: gtk-3.0-rt.stamp
	mkdir -p '$(plugins)'
	rsync -av ./gtk-3.0/ '$(plugins)/gtk-3.0/'
	touch $@
uninstall:
	rm -f $< '$(plugins)/$(app).dll'
release: all
#install release: gtk-3.0-rt.stamp
release: gtk-3.0-rt.stamp
gtk-3.0-rt.stamp: gtk-3.0/bin/gspawn-win32-helper-console.exe
gtk-3.0-rt.stamp: gtk-3.0/bin/gspawn-win32-helper.exe
gtk-3.0-rt.stamp: gtk-3.0/bin/libgtk-3-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgdk-3-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgcc_s_dw2-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libwinpthread-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libatk-1.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libcairo-gobject-2.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libcairo-2.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libepoxy-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgdk_pixbuf-2.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgio-2.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libglib-2.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgmodule-2.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgobject-2.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libintl-8.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpango-1.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpangocairo-1.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpangowin32-1.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libfontconfig-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libfreetype-6.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpixman-1-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpng16-16.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/zlib1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libexpat-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libiconv-2.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libbz2-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libharfbuzz-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpcre-1.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libgraphite2.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libffi-6.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libpangoft2-1.0-0.dll
gtk-3.0-rt.stamp: gtk-3.0/bin/libstdc++-6.dll
gtk-3.0-rt.stamp: gtk-3.0/share/icons/hicolor
gtk-3.0-rt.stamp: gtk-3.0/share/icons/Adwaita
gtk-3.0-rt.stamp:
	touch $@
#release: gtk-3.0/ieshims.dll

###############################################################################
libname=yasapi
libsources=$(filter $(libname)_%.c,$(out_yasapi_srcs))
libobjects=$(patsubst %.c,%$(EXTO),$(libsources))
libmembers=$(patsubst %,lib$(libname)$(EXTA)(%),$(libobjects))
lib$(libname)$(EXTA): $(libmembers)

headers=$(filter %.h,$(out_yasapi_srcs))
objects=$(patsubst %.c,%$(EXTO),$(filter %.c,$(out_yasapi_srcs)))
$(objects): $(headers)
.PRECIOUS: lib$(libname)$(EXTA)

###############################################################################
out_yasapi.dll: lib$(libname)$(EXTA)
x7: lib$(libname)$(EXTA)
#ifneq (ms,$(tools)) # [
$(ALL): libgtk-3-0$(EXTDLLA)
#$(ALL): libgdk-3-0$(EXTDLLA)
#$(ALL): libgobject-2.0-0$(EXTDLLA)
#$(ALL): libgio-2.0-0$(EXTDLLA)
$(ALL): libglib-2.0-0$(EXTDLLA)
#endif # ]
out_yasapi.dll: out_yasapi$(EXTO)

###############################################################################
ifeq (ms,$(tools)) # [
(%$(EXTO)): %$(EXTO)
	if test -e $@; then $(LIB) -out:$@ $@ $<; else $(LIB) -out:$@ $<; fi
%$(EXTO): %.c
	$(CC) $(CFLAGS) $(CPPFLAGS) $< -c -Fo$@
endif # ]
.PRECIOUS: %.dll
ifeq (ms,$(tools)) # [
%: %$(EXTO)
	$(CC) $(CFLAGS) $(CPPFLAGS) -Fe$@.exe $^ -link $(LDFLAGS) $(LDLIBS)
%.dll: %$(EXTO)
	$(CC) $(CFLAGS) -LD $(CPPFLAGS) -Fe$@ $^ -link $(LDFLAGS) $(LDLIBS)
.PRECIOUS: %.lib
%$(EXTDLLA): %.def
	$(LIB) -machine:x86 -def:$< -out:$@
else # ] [
%.dll: %$(EXTO)
#  $(LINK.o) -shared $^ $(LOADLIBES) $(LDLIBS) -o $@
	$(CC) $(CFLAGS) $(CPPFLAGS) $^ $(LDFLAGS) $(LDLIBS) -o $@
.PRECIOUS: %.dll.a
%$(EXTDLLA): %.def
	dlltool -d $< --output-delaylib $@
endif # ]
.PRECIOUS: %.def
#%.def:  /mingw32/bin/%.dll
%.def:  /mingw32/bin/%.dll
	gendef $<
gtk-3.0/bin/%: /mingw32/bin/%
	mkdir -p $(@D)
	cp -p $< $(@D)
gtk-3.0/share/%: /mingw32/share/%
	mkdir -p $(@D)
	cp -rp $< $(@D)
FORCE:
