/*
 * out_yasapi.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include <shlobj.h>
#include <yasapi.h>

#define YASAPI_ENGINE

#define PI_VER2 "v1.11 beta"

#ifdef __alpha
#define PI_VER PI_VER2 " (AXP)"
#else
#define PI_VER PI_VER2 " (x86)"
#endif

int getwrittentime();
int getoutputtime();

int srate, numchan, bps, active;
volatile int writtentime, w_offset;

static Player player;


BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}

Out_Module out;

#if ! defined (YASAPI_ENGINE) // [
static int last_pause=0;
#endif // ]

#define YASAI_ON_TOP_OF_PREFERENCES
#if ! defined (YASAI_ON_TOP_OF_PREFERENCES) // [
#define YASAI_ON_TOP_OF_WINAMP_MAIN
#endif // ]
void config(HWND hwnd)
{
#if defined (YASAPI_ENGINE) // [
  PlayerShow(&player,&player.gtk.config,hwnd);
#endif // ]
}

void about(HWND hwnd)
{
#if defined (YASAPI_ENGINE) // [
  PlayerShow(&player,&player.gtk.about,hwnd);
#endif // ]
}

void init()
{
	PlayerCreate(&player,&out);
}

void quit()
{
	PlayerDestroy(&player);
}

int start_t;

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
  DMARKLN();
	start_t=GetTickCount();
 	w_offset = writtentime = 0;
	active=1;
	numchan = numchannels;
	srate = samplerate;
	bps = bitspersamp;
#if defined (YASAPI_ENGINE) // [
  return PlayerOpen(&player,samplerate,numchannels,bitspersamp);
#else // ] [
	return 1;
#endif // ]
}

void close()
{
#if defined (YASAPI_ENGINE) // [
  DMARKLN();
  PlayerClose(&player);
  DMARKLN();
#endif // ]
}

int write(char *buf, int len)
{
#if defined (YASAPI_ENGINE) // [
  // 0 on success. Len == bytes to write (<= 8192 always). buf is straight
  //   audio data. 
  // 1 returns not able to write (yet). Non-blocking, always.
  if (PlayerWrite(&player,(BYTE *)buf,len)<0)
    return 1;
  else {
#endif // ]
	writtentime += len;
	return 0;
#if defined (YASAPI_ENGINE) // [
  }
#endif // ]
}

int canwrite()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerAvailable(&player);
#else // ] [
	if (last_pause) return 0;
	if (getwrittentime() < getoutputtime()+MulDiv(65536,1000,srate*bps*numchan/8)) return 65536;
	return 0;
#endif // ]
}

int isplaying()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerIsPlaying(&player);
#else // ] [
	return 0;
#endif // ]
}

#if defined (YASAPI_ENGINE) // [
int pause(int pause)
{
#if defined (YASAPI_DEBUG) // [
  static int count;
#endif // ]
  int p;
  
  //DMARKLN();
  DWRITELN("================================================================");
  DVWRITELN("count %d",++count);
  DWRITELN("================================================================");
  p=PlayerPause(&player,pause);
  //DVWRITELN("%d",p);

  return p;
}
#else // ] [
int pause(int pause)
{
	int t=last_pause;
	if (!last_pause && pause) { w_offset+=GetTickCount()-start_t;  writtentime=0; }
	if (last_pause && !pause) { start_t=GetTickCount(); }
	last_pause=pause;
	return t;
}
#endif // ]

void setvolume(int volume)
{
#if defined (YASAPI_ENGINE) && defined (YASAPI_OPTION_VOLUME) // [
  PlayerVolume(&player,volume);
#endif // ]
}

void setpan(int pan)
{
}

void flush(int t)
{
#if defined (YASAPI_ENGINE) // [
  PlayerFlush(&player);
#else // ] [
  w_offset=t;
  start_t=GetTickCount();
  writtentime=0;
#endif // ]
}
	
int getoutputtime()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerOutputTime(&player);
#else // ] [
	if (last_pause)
		return w_offset;
	return GetTickCount()-start_t + w_offset;
#endif // ]
}

int getwrittentime()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerWrittenTime(&player);
#else // ] [
	int t=srate*numchan,l;
	int ms=writtentime;

	if (t)
	{
	l=ms%t;
	ms /= t;
	ms *= 1000;
	ms += (l*1000)/t;

	ms/=(bps/8);

	return ms + w_offset;
	}
	else
		return ms;
#endif // ]
}

#if defined (YASAPI_BETA) // [
#define OUT_VER_U 0x11
#endif // ]
Out_Module out = {
#if defined (OUT_VER_U) // [
	OUT_VER_U,
  (char *)YASAPI_TITLE_U16
#else // ] [
	OUT_VER,
  YASAPI_TITLE
#endif // ]
	,
	65,
	0, // hmainwindow
	0, // hdllinstance
	config,
	about,
	init,
	quit,
	open,
	close,
	write,
	canwrite,
	isplaying,
	pause,
	setvolume,
	setpan,
	flush,
	getoutputtime,
	getwrittentime
};

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
	return &out;
}
