/*
 * yasapi_about.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>
#if defined (YASAPI_PRODUCTBUILD) // [
#include <ntverp.h>
#endif // ]

///////////////////////////////////////////////////////////////////////////////
// How to insert links in GtkTextView - Part 1 - Show Link?
// https://www.kksou.com/php-gtk2/sample-codes/insert-links-in-GtkTextView-Part-1-Show-Link.php
// How to insert links in GtkTextView - Part 2 - Activate Link?
// https://www.kksou.com/php-gtk2/sample-codes/insert-links-in-GtkTextView-Part-2-Activate-Link.php
// How to insert links in GtkTextView - Part 4 - Change Cursor over Link?
// https://www.kksou.com/php-gtk2/sample-codes/insert-links-in-GtkTextView-Part-4-Change-Cursor-over-Link.php
#if defined (YASAPI_ABOUT_TEXT_VIEW) // [
#define P "<p>"
#define BR "<b>"
#define BULL "&bull;"
typedef struct _Parser Parser;
#else // ] [
#define P "\n\n"
#define BR "\n"
#define BULL "&#8226;"
#endif // ]
typedef struct _Template Template;
typedef enum _ELimits ELimits;

#if defined (YASAPI_ABOUT_TEXT_VIEW) // [
struct _Parser {
  AboutPage *pPage;
  const char *sp;
  const char *rp;
  GtkTextIter iter;
  GtkTextMark *start;
  AboutLink *pLink;
};
#endif // ]

struct _Template {
  const char *header;
  const char *content;
};

static const Template gaTemplates[eAboutLimitsPagesNum]={
  { "License",
        "<a href=\"http://out-yasapi.sourceforge.net/\">"
        "Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output"
        "</a>"
        " (out_yasapi-nt)"
    BR "Copyright © 2015-2018 Peter Belkner"
        " <a href=\"mailto:pbelkner@snafu.de?subject=out_yasapi-nt\">"
        "&lt;pbelkner@snafu.de&gt;"
        "</a>"
    P "out_yasapi-nt is free software: you can redistribute it and/or modify"
        " it under the terms of the GNU General Public License as published"
        " by the Free Software Foundation, either version 3 of the License,"
        " or (at your option) any later version."
    P "out_yasapi-nt is distributed in the hope that it will be useful, but"
        " WITHOUT ANY WARRANTY; without even the implied warranty of"
        " MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
        " GNU General Public License for more details."
    P "You should have received a copy of the GNU General Public License"
        " along with out_yasapi-nt.  If not, see"
        " <a href=\"https://www.gnu.org/licenses\">&lt;"
        "https://www.gnu.org/licenses"
        "&gt;</a>." },
  { "References",
    "Just to name a few we'd like to thank (in no particular order)"
    BR BULL " the engineers at"
      " <a href=\"https://en.wikipedia.org/wiki/Nullsoft\">Nullsoft</a>"
      " who once created"
      " <a href=\"https://en.wikipedia.org/wiki/Winamp\">Winamp</a>,"
#if 0 // [
    BR BULL " "
      " <a href=\"https://twitter.com/Ed_Rich\">the lost one</a>"
      " at"
      " <a href=\"https://www.radionomy.com/\">Radionomy</a>"
      " for doing his best help"
      " <a href=\""
      "http://www.radionomygroup.com/en/companies/winamp/"
      "\">kicking the llama's ass</a>,"
#endif // ]
    BR BULL " the engineers at"
      " <a href=\"https://www.microsoft.com/\">Microsoft</a>"
      " for crafting"
      " <a href=\""
      "https://msdn.microsoft.com/en-us/"
      "library/windows/desktop/dd371455(v=vs.85).aspx"
      "\">WASAPI</a> and"
    BR BULL " for providing"
      " <a href=\"https://www.visualstudio.com/vs/older-downloads/\">"
      "community editions"
      "</a>"
      " of their compiler and tools,"
    BR BULL " the"
      " <a href=\"https://www.gtk.org/development.php\">GTK+ team</a>"
      " for creating this"
      " <a href=\"https://www.gtk.org/\">great GUI toolkit</a>,"
    BR BULL "  especially for providing a"
      " <a href=\"https://www.gtk.org/download/windows.php\">"
      "GTK+ port to Windows"
      "</a>,"
#if defined (YASAPI_REFERENCE_DEPENDENCY_WALKER) // [
    BR BULL " Steve Miller for crafting"
      " <a href=\"http://www.dependencywalker.com/\">"
      "Dependency Walker"
      "</a>,"
#endif // ]
    BR BULL " the"
      " <a href=\"https://en.wikipedia.org/wiki/MinGW\">MinGW</a> project"
      " for providing the"
      " <a href=\"https://gcc.gnu.org/\">GNU C compiler and tools</a>"
      " on Windows,"
    BR BULL " Alexey Pavlov for providing the"
      " <a href=\"http://www.msys2.org/\">MSYS2</a>"
      " environment,"
    BR BULL " Bram Moolenaar for crafting the"
      " <a href=\"http://www.vim.org/\">gvim</a>"
      " editor,"
    BR BULL
      "<a href=\"https://en.wikipedia.org/wiki/Richard_Stallman\">Richard"
      " Stallman</a> for crafting the "
      " <a href=\"https://www.gnu.org/licenses/licenses.html\">GPL</a>, and"
    BR BULL "<a href=\"https://sourceforge.net/\">SourceForge</a>"
      " for providing this great platform for publishing open source"
      " software."
    P "The project's web page is at"
      " <a href=\"http://out-yasapi.sourceforge.net/\">"
      "http://out-yasapi.sourceforge.net/"
      "</a>."
      " The download page is at"
      " <a href=\"https://sourceforge.net/projects/out-yasapi/files/\">"
      "https://sourceforge.net/projects/out-yasapi/files/yasapi-nt"
      "</a>." },
  { "Notes",
    "MERKEL MUß WEG! JETZT! GENUG IST GENUG!"
    BR "Merkel has to leave office! Right now! Enough is enough!"
    P "<a href=\""
      "https://en.wikipedia.org/wiki/Standing_on_the_shoulders_of_giants"
      "\">Nanos gigantum humeris insidentes:"
      "</a>"
      " This project is dedicated to my"
      " European heritage. It is strictly to be understood as a statement"
      " against the \"sweet\" liberal lie of \"multiculturalism\" which"
      " is going to destroy Europe as we know it, in particular against"
      " the Merkel regime selling out Europe for nothing as we watch."
    P "PLEASE NOTE THAT THIS PROJECT IS AN EXPERIMENTAL RATHER THEN AN"
      " INDUSTRIAL STRENGTH EFFORT. THIS PROJECT IS NOT FOR YOU. IT IS FOR"
      " ME IN ORDER TO LEARN SOMETHING. IF THERE IS SOMETHING ALONG THE"
      " WAY I CAN DO FOR YOU THAT'S GREAT!"
    P "Peter Belkner"
      " <a href=\"http://home.snafu.de/pbelkner/\">"
      "&lt;http://home.snafu.de/pbelkner/&gt;"
      "</a>" },
};

#if defined (YASAPI_ABOUT_TEXT_VIEW) // [
enum _ELimits {
  ePagesNum=(sizeof gaPages)/(sizeof gaPages[0]),
  eWindowWidthPixels=455,
};

static gboolean window_delete_event (GtkWidget *window, GdkEvent *event,
    gpointer user_data)
{
  int width = gtk_widget_get_allocated_width (window);
  int height = gtk_widget_get_allocated_height (window);

  fprintf (stderr, "%s:%d: width=%d height=%d\n", __func__, __LINE__,
      width, height);
  fflush (stderr);

  gtk_widget_hide (window);

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
static void AboutLinkApplyTag(AboutLink *pLink, GtkTextBuffer *buffer,
    GtkTextTag *remove, GtkTextTag *apply)
{
  GtkTextIter start,end;

  gtk_text_buffer_get_iter_at_mark (buffer, &start, pLink->start.mark);
  gtk_text_buffer_get_iter_at_mark (buffer, &end, pLink->end.mark);

  if (remove)
    gtk_text_buffer_remove_tag (buffer, remove, &start, &end);

  if (apply)
    gtk_text_buffer_apply_tag (buffer, apply, &start, &end);
}

#define NOTEBOOK_SWITCH_PAGE
#if defined (NOTEBOOK_SWITCH_PAGE) // [
static void notebook_switch_page (GtkNotebook *notebook, GtkWidget *page,
    guint page_num, gpointer user_data)
{
  About *pAbout = user_data;
  AboutPage *pPageActive;
  AboutLink *pLinkActive;
  GtkTextTag *remove,*apply;

  if (!pAbout->pPageActive)
    goto page;

  pPageActive = pAbout->pPageActive;

  if (!pPageActive->pLinkActive)
    goto link;

  pLinkActive = pPageActive->pLinkActive;

  if (!pPageActive->buffer)
    goto buffer;

  remove = pLinkActive->anchor->bold;
  apply = pLinkActive->anchor->normal;
  AboutLinkApplyTag(pLinkActive, pPageActive->buffer, remove, apply);
buffer:
  pPageActive->pLinkActive = NULL;
link:
page:
  if (0 <= page_num && page_num < eAboutLimitsPagesNum) {
    pAbout->pPageActive = pAbout->pages + page_num;
    // TODO: Out of nowhere pLinkActive might point to some invalid location.
    pAbout->pPageActive->pLinkActive = NULL;
  }
  else
    pAbout->pPageActive = NULL;
}
#endif // ]

#define BUFFER_MARK_SET
#if defined (BUFFER_MARK_SET) // [
static void buffer_mark_set (GtkTextBuffer *buffer, GtkTextIter *location,
    GtkTextMark *mark, gpointer user_data)
{
  About *pAbout = user_data;
  AboutPage *pPageActive;
  int offs;
  const char *name;
#if defined (ABOUT_RESTORE_CURSOR_HAND) // [
  GdkWindow *window;
#endif // ]
  AboutLink *pLink;
  GtkTextTag *remove,*apply;

  pPageActive = pAbout->pPageActive;
  offs=gtk_text_iter_get_offset (location);
  name = gtk_text_mark_get_name (mark);
#if defined (ABOUT_RESTORE_CURSOR_HAND) // [
  window = gtk_text_view_get_window (GTK_TEXT_VIEW (pPageActive->view),
        GTK_TEXT_WINDOW_TEXT);
#endif // ]

  if (name && ! strcmp ("insert", name)) {
    pLink = pPageActive->links;

    while (pLink->url[0]) {
      if (offs < pLink->start.offs)
        break;
      else if (pLink->end.offs <= offs)
        ++pLink;
      else {
        fprintf(stderr, "mark-set: \"%s\" %d [%d, %d) \"%s\"\n", name, offs,
            pLink->start.offs, pLink->end.offs, pLink->url);
        fflush (stderr);

        ShellExecuteA(
          NULL,           // _In_opt_ HWND    hwnd,
          "open",         // _In_opt_ LPCTSTR lpOperation,
          pLink->url,     // _In_     LPCTSTR lpFile,
          NULL,           // _In_opt_ LPCTSTR lpParameters,
          NULL,           // _In_opt_ LPCTSTR lpDirectory,
          SW_SHOWNORMAL   // _In_     INT     nShowCmd
        );

        remove=pLink->anchor->bold;
        pLink->anchor=&pPageActive->black;
        apply=pLink->anchor->normal;

        if (remove != apply)
          AboutLinkApplyTag(pLink, pPageActive->buffer, remove, apply);
#if defined (ABOUT_RESTORE_CURSOR_HAND) // [
        gdk_window_set_cursor (window, pAbout->hand);
#endif // ]
        break;
      }
    }
  }
}
#endif // ]

#define VIEW_MOTION_NOTIFY_EVENT
#if defined (VIEW_MOTION_NOTIFY_EVENT) // [
static gboolean view_motion_notify_event (GtkWidget *view, GdkEvent *event,
    gpointer user_data)
{
  enum { eDelta = 4 };
  About *pAbout = user_data;
  AboutPage *pPageActive;
  AboutLink *pLink;
  GtkTextTag *remove,*apply;
  GdkWindow *window;
  gint x,y;
  int width,height;
  GtkTextIter iter;
  int offs;

  if (!pAbout->pPageActive)
    goto exit;

  pPageActive = pAbout->pPageActive;

  width = gtk_widget_get_allocated_width (view);
  height = gtk_widget_get_allocated_height (view);
  window = gtk_text_view_get_window (GTK_TEXT_VIEW (view),
      GTK_TEXT_WINDOW_TEXT);
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
      GTK_TEXT_WINDOW_TEXT, event->motion.x, event->motion.y, &x, &y);

  if (x<=eDelta || y<=eDelta || (width-1-eDelta)<=x || (height-1-eDelta)<=y) {
    // We're about to leave.
    pLink = pPageActive->pLinkActive;

    if (pLink) {
      remove = pLink->anchor->bold;
      apply = pLink->anchor->normal;
      AboutLinkApplyTag(pLink, pPageActive->buffer, remove, apply);
      pPageActive->pLinkActive = NULL;
    }

    gdk_window_set_cursor (window, pAbout->arrow);
    goto exit;
  }

  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (view), &iter, x, y);
  offs = gtk_text_iter_get_offset (&iter);

  pLink = pPageActive->pLinkActive;

  if (pLink) {
    if (pLink->start.offs <= offs && offs < pLink->end.offs)
      goto exit;
    else {
      remove = pLink->anchor->bold;
      apply = pLink->anchor->normal;
      AboutLinkApplyTag(pLink, pPageActive->buffer, remove, apply);
      pPageActive->pLinkActive = NULL;
    }
  }

  pLink = pPageActive->links;

  while (pLink->url[0]) {
    if (offs < pLink->start.offs)
      break;
    else if (pLink->end.offs <= offs)
      ++pLink;
    else {
      remove = pLink->anchor->normal;
      apply = pLink->anchor->bold;
      AboutLinkApplyTag(pLink, pPageActive->buffer, remove, apply);
      gdk_window_set_cursor (window, pAbout->hand);
      pPageActive->pLinkActive=pLink;
      goto exit;
    }
  }

  gdk_window_set_cursor (window, pAbout->arrow);
exit:
  return FALSE;
}
#endif // ]

static void ParserFlush (Parser *pParser, const char *sp, const char *rp)
{
  GtkTextBuffer *buffer=pParser->pPage->buffer;
  GtkTextIter *iter=&pParser->iter;
  int n=rp?rp-sp:-1;

  gtk_text_buffer_insert(buffer,iter,sp,n);
}

static void entity (Parser *pParser, const char *tag)
{
  // Remeber entitie's start.
  const char *ep = pParser->rp;

  while (';'!=*pParser->rp)
    ++pParser->rp;

  // https://brajeshwar.github.io/entities/
  if (! memicmp (ep, "lt", 2))
    ParserFlush (pParser, "<", NULL);
  else if (! memicmp (ep, "gt", 2))
    ParserFlush (pParser, ">", NULL);
  else if (! memicmp (ep, "bull", 4) || ! memcmp (ep, "#8226", 5))
    ParserFlush (pParser, "\u2022", NULL);
  else
    ParserFlush (pParser, ep-1, pParser->rp);

  // Consume ';'.
  ++pParser->rp;
  pParser->sp = pParser->rp;
}

//#define FLUSH_IN_TAG_MARKER
static void tag_start (Parser *pParser, char *tag, char *val)
{
  GtkTextBuffer *buffer=pParser->pPage->buffer;
  const char *tp,*vp;
  int n;

#if defined (FLUSH_IN_TAG_MARKER) // [
  // Flush.
  // rp is at '<tag> ...'
  //            ^
  ParserFlush (pParser, pParser->sp, pParser->rp - 1);
#endif // ]

  // Remeber start of tag.
  tp=pParser->rp;
  tag[0]=0;

  while ('>' != *pParser->rp) {
    if (' ' == *pParser->rp) {
      n = pParser->rp - tp;
      memcpy(tag, tp, n);
      tag[n] = 0;

      // Consume ' '.
      ++pParser->rp;

      while ('"' != *pParser->rp)
        ++pParser->rp;
              
      // Consume '"'.
      ++pParser->rp;

      // Remeber the start of value.
      vp = pParser->rp;

      while ('"' != *pParser->rp)
        ++pParser->rp;
              
      n = pParser->rp - vp;
      memcpy (val, vp, n);
      val[n] = 0;
fprintf (stderr, "val=\"%s\"\n", val); fflush (stderr);

      // Consume '"'.
      ++pParser->rp;
    }
    else
      ++pParser->rp;
  }

  if (!tag[0]) {
    n = pParser->rp - tp;
    memcpy(tag, tp, n);
    tag[n] = 0;
  }

  if (! stricmp ("a", tag)) {
    pParser->start = gtk_text_buffer_create_mark (buffer, NULL,
        &pParser->iter, TRUE);
  }

  // Consume '>'.
  ++pParser->rp;
  pParser->sp = pParser->rp;
}

static void AboutBoundSet(AboutBound *pBound, GtkTextBuffer *buffer,
    GtkTextMark *mark)
{
  GtkTextIter iter;

  pBound->mark = mark;
  gtk_text_buffer_get_iter_at_mark (buffer, &iter, mark);
  pBound->offs = gtk_text_iter_get_offset (&iter);
}

static void tag_end (Parser *pParser, char *tag, const char *val)
{
  GtkTextBuffer *buffer=pParser->pPage->buffer;
  AboutLink *pLink = pParser->pLink;
  GtkTextMark *end;

#if defined (FLUSH_IN_TAG_MARKER) // [
  // Flush.
  // rp is at '</tag> ...'
  //            ^
  ParserFlush (pParser, pParser->sp, pParser->rp - 1, tag);
#endif // ]

  if (! stricmp ("a", tag)) {
    pLink->pPage = pParser->pPage;
    strcpy (pParser->pLink->url, val);

    end = gtk_text_buffer_create_mark (buffer, NULL, &pParser->iter, TRUE);

    AboutBoundSet(&pLink->start, buffer, pParser->start);
    AboutBoundSet(&pLink->end, buffer, end);
    pLink->anchor=&pParser->pPage->blue;
    AboutLinkApplyTag(pLink, buffer, NULL, pLink->anchor->normal);

    pParser->start = NULL;
    ++pParser->pLink;
  }

  // Consume '/'.
  ++pParser->rp;

  while ('>' != *pParser->rp)
    ++pParser->rp;

  // Consume '>'.
  ++pParser->rp;
  pParser->sp = pParser->rp;
}

//#define PARSE_RECURSE
#if defined (PARSE_RECURSE) // [
static void parse (Parser *pParser, int depth)
#else // ] [
static void parse (Parser *pParser)
#endif // ]
{
  GtkTextBuffer *buffer=pParser->pPage->buffer;
  GtkTextIter *iter=&pParser->iter;
  char tag[32]={0},val[128]={0};

  while (*pParser->rp) {
    if ('<'==*pParser->rp) {
      // Consume '<'.
      ++pParser->rp;

      if ('/'==*pParser->rp) {
        // End tag, i.e. </tag>.
#if defined (PARSE_RECURSE) // [
        if (depth) {
          --pParser->rp;
          return;
        }
#endif // ]
#if ! defined (FLUSH_IN_TAG_MARKER) // [
        // Flush.
        // rp is at '</ ... >'
        //            ^
        ParserFlush (pParser, pParser->sp, pParser->rp - 1);
#endif // ]
        tag_end (pParser, tag, val);
      }
      else {
        // Start tag, i.e. <tag>.
#if ! defined (FLUSH_IN_TAG_MARKER) // [
        // Flush.
        // rp is at '< ... >'
        //            ^
        ParserFlush (pParser, pParser->sp, pParser->rp - 1);
#endif // ]
        tag_start (pParser, tag, val);

        if (!strcmp(tag, "br"))
          gtk_text_buffer_insert(buffer, iter, "\n", -1);
        else if (!strcmp(tag, "p"))
          gtk_text_buffer_insert(buffer, iter, "\n\n", -1);
#if defined (PARSE_RECURSE) // [
        else {
          // Recurse.
          parse (pParser, ++depth);
        }
#endif // ]
      }
    }
    else if ('&' == *pParser->rp) {
      // Flush.
      // rp is at '& ... ;'
      //           ^
      ParserFlush (pParser, pParser->sp, pParser->rp);

      // Consume '&'.
      ++pParser->rp;
      entity (pParser, tag);
    }
    else
      ++pParser->rp;
  }
}

static GtkWidget *yasapi_about_new (Dialog *pDialog)
{
  GtkWidget *window,*vbox,*branch,*leaf,*helper;
  const Template *pTemplate=gaTemplates;
  const Template *pTemplateMax=gaTemplates+eAboutLimitsPagesNum;
  About *pAbout=&pDialog->about;
  AboutPage *pPage=pAbout->pages;
  Parser parser;
  GdkRGBA blue;
  GdkRGBA black;

  gdk_rgba_parse (&blue, "rgb(42,118,198)");
  gdk_rgba_parse (&black, "rgb(0,0,0)");

  /////////////////////////////////////////////////////////////////////////////
  pAbout->display = gdk_display_get_default ();
fprintf(stderr, "display=%p\n", pAbout->display); fflush (stderr);
  pAbout->arrow = gdk_cursor_new_for_display (pAbout->display, GDK_ARROW);
fprintf(stderr, "arrow=%p\n", pAbout->arrow); fflush (stderr);
  pAbout->hand = gdk_cursor_new_for_display (pAbout->display, GDK_HAND1);
fprintf(stderr, "hand=%p\n", pAbout->hand); fflush (stderr);

  // Build the widget tree. ///////////////////////////////////////////////////
  // window
  // +- vbox
  //    +- notebook
  //       +- notebook page*
  //          +- label
  //          +- view
  //    +- button_box
  //       +- button
  // window.
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
  gtk_window_set_default_size (GTK_WINDOW (window),
      eAboutLimitsWidthPixels, -1);
  gtk_window_set_title (GTK_WINDOW (window), "About " YASAPI_TITLE_U8);
  gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
#if 0 // [
  g_signal_connect(G_OBJECT(window), "destroy",
      G_CALLBACK(gtk_main_quit), "window");
#endif // ]
  g_signal_connect(G_OBJECT(window), "delete-event",
      G_CALLBACK(window_delete_event), NULL);

  // vbox
  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  // notebook
  pDialog->notebook = branch = gtk_notebook_new ();
  gtk_container_add (GTK_CONTAINER (vbox), branch);

  while (pTemplate<pTemplateMax) {
    pPage->pAbout = pAbout;
    parser.pPage = pPage;
    // label.
    helper = gtk_label_new (pTemplate->header);

    // text_view.
    leaf = gtk_text_view_new ();
#if defined (ABOUT_RESTORE_CURSOR_HAND) // [
    pPage->view = leaf;
#endif // ]
#if defined (VIEW_MOTION_NOTIFY_EVENT) // [
    g_signal_connect(G_OBJECT(leaf), "motion-notify-event",
        G_CALLBACK(view_motion_notify_event), pAbout);
#endif // ]
    gtk_widget_set_valign (leaf, GTK_ALIGN_CENTER);
    gtk_widget_set_margin_top (leaf, ePanelSpace<<1);
    gtk_widget_set_margin_bottom (leaf, ePanelSpace<<1);
    gtk_widget_set_margin_start (leaf, ePanelSpace<<1);
    gtk_widget_set_margin_end (leaf, ePanelSpace<<1);
    gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (leaf), GTK_WRAP_WORD);
    gtk_text_view_set_editable (GTK_TEXT_VIEW (leaf), FALSE);
    gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (leaf), FALSE);

    memset (pPage->links, 0, sizeof pPage->links);
    parser.pLink = pPage->links;
    parser.rp = parser.sp = pTemplate->content;
    pPage->buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(leaf));
    pPage->blue.normal = gtk_text_buffer_create_tag (pPage->buffer, NULL,
        "foreground-rgba", &blue,
        "weight", PANGO_WEIGHT_NORMAL,
        "underline", PANGO_UNDERLINE_SINGLE,
        NULL); 
    pPage->blue.bold = gtk_text_buffer_create_tag (pPage->buffer, NULL,
        "foreground-rgba", &blue,
#if defined (SEMIBOLD) // [
        "weight", PANGO_WEIGHT_SEMIBOLD,
#else // ] [
        "weight", PANGO_WEIGHT_BOLD,
#endif // ]
        "underline", PANGO_UNDERLINE_SINGLE,
        NULL); 
    pPage->black.normal = gtk_text_buffer_create_tag (pPage->buffer, NULL,
        "foreground-rgba", &black,
        "weight", PANGO_WEIGHT_NORMAL,
        "underline", PANGO_UNDERLINE_SINGLE,
        NULL); 
    pPage->black.bold = gtk_text_buffer_create_tag (pPage->buffer, NULL,
        "foreground-rgba", &black,
#if defined (SEMIBOLD) // [
        "weight", PANGO_WEIGHT_SEMIBOLD,
#else // ] [
        "weight", PANGO_WEIGHT_BOLD,
#endif // ]
        "underline", PANGO_UNDERLINE_SINGLE,
        NULL); 
    //gtk_text_buffer_get_iter_at_offset(pPage->buffer, &parser.iter, 0);
    gtk_text_buffer_get_end_iter(pPage->buffer, &parser.iter);
#if defined (PARSE_RECURSE) // [
    parse (&parser, 0);
#else // ] [
    parse (&parser);
#endif // ]
    gtk_text_buffer_insert(pPage->buffer, &parser.iter, parser.sp, -1);
#if defined (BUFFER_MARK_SET) // [
    g_signal_connect (pPage->buffer, "mark-set", G_CALLBACK (buffer_mark_set),
        pAbout);
#endif // ]
    gtk_notebook_append_page (GTK_NOTEBOOK (branch), leaf, helper);
    ++pPage;
    ++pTemplate;
  }

#if defined (NOTEBOOK_SWITCH_PAGE) // [
  g_signal_connect(G_OBJECT(branch), "switch-page",
      G_CALLBACK(notebook_switch_page), pAbout);
#endif // ]

  // button_box
  branch = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
  gtk_widget_set_margin_top (branch, ePanelSpace);
  gtk_widget_set_margin_bottom (branch, ePanelSpace);
  gtk_container_add (GTK_CONTAINER (vbox), branch);

  // button "Ok".
  leaf = gtk_button_new_with_label ("Ok");
  gtk_container_add (GTK_CONTAINER (branch), leaf);
  g_signal_connect_swapped (leaf, "clicked",
      G_CALLBACK (gtk_widget_hide), window);

  return window;
}
#else // ] [
static int config_glib_version(Dialog *pDialog, char *buf, size_t size)
{
  // Delayloading the "gtk_<xxx>_version" DATA symbols seem not to work,
  // hence we load them directly from the respective DLL.
  HMODULE hLib=pDialog->pGtk->hLibGlib;
  guint major=*(guint *)GetProcAddress(hLib,"glib_major_version");
  guint minor=*(guint *)GetProcAddress(hLib,"glib_minor_version");
  guint micro=*(guint *)GetProcAddress(hLib,"glib_micro_version");

  return sprintf(buf,"%d.%d.%d",major,minor,micro);
}

static int config_gtk_version(Dialog *pDialog, char *buf, size_t size)
{
  // The "gtk_<xxx>_version" symbols are macro definitions on top of the
  // respective "gtk_get_<xxx>_version()" functions coming from the CODE
  // segment.
  gint major=gtk_major_version;
  gint minor=gtk_minor_version;
  gint micro=gtk_micro_version;

  return sprintf(buf,"%d.%d.%d",major,minor,micro);
}

// Needs helper application in order to open links:
// https://github.com/mypaint/mypaint/issues/551#issuecomment-210572560
static void AboutWindowCreate(Dialog *pDialog)
{
  ////////////////
#if defined (_MSC_VER) // [
  static char aMscVer[(sizeof YASAPI_STRA(_MSC_FULL_VER))+2]
      =YASAPI_STRA(_MSC_FULL_VER);
#endif // ]

  static ClassPage version_rows={
    { LABEL_STATIC(2,0,"<b>Statically</b>",NULL) },
    { LABEL_STATIC(2,0,"(at Compile Time on <i>Our</i> System)",NULL) },
    { LABEL_STATIC(1,0,"YASAPI/NT:",NULL),
        LABEL_STATIC(1,0,YASAPI_VERSION_U8 YASAPI_TITLE_SFX_U8,NULL) },
#if defined (_MSC_VER) // [
    { LABEL_STATIC(1,0,"msc:",NULL),
        LABEL_STATIC(1,0,aMscVer,NULL) },
#elif defined (__GNUC__) // ] [
    { LABEL_STATIC(1,0,"gcc:",NULL),
        LABEL_STATIC(1,0,YASAPI_STRA(__GNUC__)
        "." YASAPI_STRA(__GNUC_MINOR__)
        "." YASAPI_STRA(__GNUC_PATCHLEVEL__),NULL) },
#endif // ]
#if defined (YASAPI_PRODUCTBUILD) && defined (VER_PRODUCTBUILD) // [
    { LABEL_STATIC(1,0,"win sdk:",NULL),
        LABEL_STATIC(1,0,YASAPI_STRA(VER_PRODUCTBUILD),NULL) },
#endif // ]
    { LABEL_STATIC(2,0,"",NULL) },
    { LABEL_STATIC(2,0,"<b>Dynamically</b>",NULL) },
    { LABEL_STATIC(2,0,"(Loaded at Runtime from <i>Your</i> System)",NULL) },
    { LABEL_STATIC(1,0,"libglib:",NULL),
        LABEL_STATIC_FUN(1,0,config_glib_version) },
    { LABEL_STATIC(1,0,"libgtk:",NULL),
        LABEL_STATIC_FUN(1,0,config_gtk_version) },
    { { NULL } }
  };

  static ClassCell cell=PAGE(version_rows);

  GtkWidget *window,*vbox,*branch,*leaf,*helper;
  const Template *pTemplate=gaTemplates;
  const Template *pTemplateMax=gaTemplates+eAboutLimitsPagesNum;
  Expand e;

#if defined (_MSC_VER) // [
  if ('.'!=aMscVer[2]) {
    memmove(aMscVer+3,aMscVer+2,(sizeof YASAPI_STRA(_MSC_FULL_VER))-2);
    aMscVer[2]='.';
    memmove(aMscVer+6,aMscVer+5,(sizeof YASAPI_STRA(_MSC_FULL_VER))-4);
    aMscVer[5]='.';
  }
#endif // ]

  // Build the widget tree. ///////////////////////////////////////////////////
  // window
  // +- vbox
  //    +- notebook
  //       +- notebook page*
  //          +- label
  //          +- label
  //    +- button_box
  //       +- button
  // window.
  pDialog->window = window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  // vbox
  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  // notebook
  pDialog->notebook = branch = gtk_notebook_new ();
  g_object_set (G_OBJECT (branch), "margin", ePanelSpace, NULL);
  gtk_container_add (GTK_CONTAINER (vbox), branch);

  while (pTemplate<pTemplateMax) {
    // label.
    helper = gtk_label_new (pTemplate->header);

    // page.
    leaf = gtk_label_new (NULL);
    g_object_set (G_OBJECT (leaf), "margin", ePanelSpace<<1, NULL);
    gtk_label_set_use_markup (GTK_LABEL (leaf), TRUE);
    gtk_label_set_markup (GTK_LABEL (leaf),  pTemplate->content);
    gtk_label_set_line_wrap (GTK_LABEL (leaf), TRUE);
    gtk_label_set_width_chars (GTK_LABEL (leaf), eAboutLimitsWidthChars);
    gtk_notebook_append_page (GTK_NOTEBOOK (branch), leaf, helper);

    ++pTemplate;
  }

  // versions label.
  helper = gtk_label_new ("Versions");

  // versions page.
  e.h = FALSE;
  e.v = FALSE;
  leaf = cell.lpVtbl->create (&cell,pDialog,&e);
  gtk_widget_set_halign (leaf, GTK_ALIGN_CENTER);
  gtk_widget_set_valign (leaf, GTK_ALIGN_CENTER);
  gtk_notebook_append_page (GTK_NOTEBOOK (branch), leaf, helper);

  // button_box
  branch = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
  gtk_widget_set_margin_bottom (branch, ePanelSpace);
  gtk_container_add (GTK_CONTAINER (vbox), branch);

  // button "Ok".
  leaf = gtk_button_new_with_label ("Ok");
  gtk_container_add (GTK_CONTAINER (branch), leaf);
  g_signal_connect (leaf, "clicked", G_CALLBACK (gtk_dialog_options_save),
      pDialog);
  g_signal_connect_swapped (leaf, "clicked",
      G_CALLBACK (gtk_widget_hide), window);
  gtk_widget_set_tooltip_text(leaf, "Hide the window.");
}
#endif // ]

static void AboutSet(Dialog *pDialog)
{
}

static void AboutGet(Dialog *pDialog)
{
}

IDialogVtbl gIAboutVtbl={
#if defined (YASAPI_DEBUG) // [
  .pIdA="About",
#endif // ]
  .pIdW=L"About",
  .pTitle="About " YASAPI_TITLE_U8,
  .nWidthPixels=eAboutWidthPixels,
  .WindowCreate=AboutWindowCreate,
  .set=AboutSet,
  .get=AboutGet,
};
