/*
 * yasapi_config.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
const char CONFIG_WIDGET[]="ConfigWidget";

///////////////////////////////////////////////////////////////////////////////
char *NavigatePlayer(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;

  return ((char *)pPlayer)+offs;
}

char *NavigateDeviceSelectedConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  WasapiDevice *pWasapiDevice=pWasapi->selected.pDeviceConfig;

  return ((char *)pWasapiDevice)+offs;
}

char *NavigateOptionsCommonConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  OptionsCommon *pOptionsCommon=&pWasapi->options.common.config;

  return ((char *)pOptionsCommon)+offs;
}

char *NavigateOptionsDeviceConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  WasapiDevice *pWasapiDevice=pWasapi->selected.pDeviceConfig;
  OptionsDevice *pOptionsDevice=&pWasapiDevice->pMember->config;

  return ((char *)pOptionsDevice)+offs;
}

char *NavigateEngine(Dialog *pDialog, int offs)
{
  Engine *pEngine=&pDialog->pGtk->pPlayer->engine;

  return eEngineNull<pEngine->eState?((char *)pEngine+offs):NULL;
}

///////////////////////////////////////////////////////////////////////////////
ConfigWidget *ConfigWidgetCreate(ClassCell *pCell, Dialog *pDialog,
    IConfigWidgetVtbl *lpVtbl, GtkWidget *widget)
{
  ConfigWidget *pWidget;

  if (0<=pCell->widget&&lpVtbl) {
    pWidget = (ConfigWidget *)((char *)&pDialog->config + pCell->widget);
    pWidget->lpVtbl=lpVtbl;
    pWidget->pDialog=pDialog;
    pWidget->pCell=pCell;
    pWidget->widget=widget;
    g_object_set_data(G_OBJECT(widget), CONFIG_WIDGET, pWidget);
    // It's not a good idea to call "set()" on our individual widget
    // because it may have side effects to other, not yet initialized
    // widgets.
    // That's the theory. In practice we need both.
    lpVtbl->set(pWidget);

#if ! defined (YASAPI_OPTION_VOLUME) // [
    if (pWidget==&pDialog->config.common.volume)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
#if ! defined (YASAPI_OPTION_GAPLESS) // [
    if (pWidget==&pDialog->config.common.gapless)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
/*
    if (pWidget==&pDialog->config.yasapi.bitdepth)
      gtk_widget_set_sensitive (widget,0);
*/
#if 0 && ! defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    if (pWidget==&pDialog->config.wasapi.pro_audio)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
  }
  else
    pWidget = NULL;

  if (pCell->tooltip)
    //gtk_widget_set_tooltip_text(widget, pCell->tooltip);
    gtk_widget_set_tooltip_markup (widget, pCell->tooltip);

  return pWidget;
}

static void ConfigWindowCreate(Dialog *pDialog)
{
  /////////////////////////////////////////////////////////////////////////////
  static ConfigComboBoxMap mode_map[]={
    { "Shared", "<i>Shared</i>", eModeShared },
    { "Exclusive", "<i>Exclusive</i>", eModeExclusive },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap strategy_map[]={
    { "Push", "<i>Push</i>", eStrategyPush },
    { "Pull", "<i>Pull</i>", eStrategyPull },
    { NULL, NULL, -1 }
  };

  static const char attDevice[]="Choose a <i>WASAPI</i> enabled device.";

  static const char attMode[]=
    "Whether the plug-in should operate with <i>WASAPI</i> mode"
      " <i>Exclusive</i> or <i>Shared</i>."
    ;
  static const char attStrategy[]=
    "Whether the plug-in should operate with <i>WASAPI</i> strategy"
      " <i>Push</i> or <i>Pull</i>."
    ;

  static ClassPage general_rows = {
    { LABEL_STATIC(1,0,"Device:",attDevice),
        COMBO_BOX_DEVICE_WASAPI(1,0,general.device,attDevice) },
    { LABEL_STATIC(1,0,"Mode:",attMode), COMBO_BOX_DEVICE(1,0,
        general.mode,eMode,mode_map,attMode) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy), COMBO_BOX_DEVICE(1,0,
        general.strategy,eStrategy,strategy_map,attStrategy) },
    { { NULL } }
  };

  ////////////////
  static const char attLatency[]=
      "Whether <i>YASAPI/NT</i> should optimize getoutputtime() and"
      " getwrittentime() by not using IAudioClock and always return zero"
      " and the latency resulting from <i>YASAPI/NT</i>'s buffer,"
      " respectively."
      ;
  static const char attMono2Stereo[]=
      "Duplicate the only channel from mono to a second channel for (pseudo)"
      " stereo (2.0)."
      ;
  static const char attVolume[]=
      "Enable volume control."
#if ! defined (YASAPI_OPTION_VOLUME) // [
      " Not implemented yet. Always disabled."
#endif // ]
      ;
  static const char attSurround[]="How <i>4</i>, <i>6</i>, and <i>8</i>"
      " channels shoud be interpreted: as <i>surround</i>,"
      " <i>5.1 surround</i>, and <i>7.1 surround</i> (enabled),"
      " or as <i>quadrophonic</i>, <i>5.1</i>, and <i>7.1</i> (disabled)."
      ;
  static const char attDisconnect[]="Wether <i>YASAPI/NT</i> should"
      " release it's instances of IAudioClient, IAudioRenderClient, and"
      " IAudioClock during pause and re-initialize them when done."
      ;
  static const char attGapless[]=
      "Enable gapless playback."
#if ! defined (YASAPI_OPTION_VOLUME) // [
      " Not implemented yet. Always disabled."
#endif // ]
      ;
  static const char attConfigPos[]=
    "Where the \"configuration\" dialog should be positioned."
    ;
  static const char attAboutPos[]=
    "Where the \"about\" dialog should be positioned."
    ;

  static ConfigComboBoxMap mono2stero_map[]={
    { "Off","<i>Off",eMono2StereoOff },
    { "Duplicate Plain","<i>Duplicate Plain</i>",eMono2StereoPlain },
    { "Duplicate Mutiplied by ½","<i>Duplicate Mutiplied by ½</i>",
        eMono2StereoHalf },
    { "Duplicate Mutiplied by 1/√2","<i>Duplicate Mutiplied by 1/√2</i>",
        eMono2StereoSqrt2Inverse },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap pos_map[]={
    { "Free Floating","<i>Free Floating</i>",eDialogPosFreeFloating },
    { "On Top of Preferences","<i>On Top of Preferences</i>",
        eDialogPosOnTopPreferences },
    { "Always on Top","<i>Always on Top</i>",eDialogPosAlwaysOnTop },
    { NULL, NULL, -1 }
  };

#if defined (YASAPI_VOLUME_WASAPI) // [
  static ConfigComboBoxMap volume_map[]={
    { "Off","<i>Off</i>",eVolumeOff },
    { "YASAP/NT","<i>YASAP/NT</i>",eVolumeSoftware },
#if defined (YASAPI_SIMPLE_AUDIO_VOLUME) \
      && defined (YASAPI_AUDIO_ENDPOINT_VOLUME) // [
    { "WASAPI","<i>WASAPI</i>",eVolumeWasapi },
#elif defined (YASAPI_SIMPLE_AUDIO_VOLUME) // ] [
    { "IAudioSimpleVolume","<i>IAudioSimpleVolume</i>",eVolumeWasapi },
#else // ] [
    { "IAudioEndpointVolume","<i>IAudioEndpointVolume</i>",eVolumeWasapi },
#endif // ]
    { NULL, NULL, -1 }
  };
#endif // ]

  static ClassPage common_rows={
    { LABEL_STATIC(1,0,"Report Latency Instead of Time:",attLatency), 
        CHECK_BUTTON_COMMON(1,0,common.latency,bLatency,attLatency) },
    { LABEL_STATIC(1,0,"Play Mono as Stereo:",attMono2Stereo), 
        COMBO_BOX_COMMON(1,0,common.mono2stereo,eMono2Stereo,
            mono2stero_map,attMono2Stereo) },
#if defined (YASAPI_VOLUME_WASAPI) // [
    { LABEL_STATIC(1,0,"Volume Control:",attVolume), 
        COMBO_BOX_COMMON(1,0,common.volume,eVolume,volume_map,attVolume) },
#else // ] [
    { LABEL_STATIC(1,0,"Volume Control:",attVolume), 
        CHECK_BUTTON_COMMON(1,0,common.volume,eVolume,attVolume) },
#endif // ]
    { LABEL_STATIC(1,0,"Surround:",attSurround), 
        CHECK_BUTTON_COMMON(1,0,common.surround,bSurround,attSurround) },
    { LABEL_STATIC(1,0,"Disconnect when Pause:",attDisconnect), 
        CHECK_BUTTON_COMMON(1,0,common.disconnect,bDisconnect,attDisconnect) },
    { LABEL_STATIC(1,0,"Gapless Playback:",attGapless), 
        CHECK_BUTTON_COMMON(1,0,common.gapless,bGapless,attGapless) },
    { LABEL_STATIC(1,0,"Configuration Dialog Position:",attConfigPos), 
        COMBO_BOX_COMMON(1,0,common.pos.config,pos.eConfig,pos_map,
            attConfigPos) },
    { LABEL_STATIC(1,0,"About Dialog Position:",attAboutPos), 
        COMBO_BOX_COMMON(1,0,common.pos.about,pos.eAbout,pos_map,
            attAboutPos) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap period_map[]={
    { "Default", "<i>Default</i>", ePeriodDefault },
    { "Minimum", "<i>Minimum</i>", ePeriodMinimum },
    { NULL }
  };

  static const char attAutoConvertPCM[]=
      "Whether <i>WASAPI</i> should automatically convert to the desiered"
      " format. Available only with mode <i>Shared</i>. Disabled otherwise."
      ;
  static const char attSRCDefaultQuality[]=
      "Whether (according to <i>Microsoft</i>) \"a sample rate converter with"
      " better quality than the default conversion but with a higher"
      " performance cost is used. This should be used if the audio is"
      " ultimately intended to be heard by humans as opposed to other"
      " scenarios such as pumping silence or populating a meter.\""
      ;
  static const char attTimerPeriodic[]=
    "<i>Microsoft</i> waitable timers are not strictly <i>WASAPI</i>"
    " objects. <i>YASAPI/NT</i> uses one in order to implement strategy"
    " <i>Push</i> and the option fits perfectly on this page. If enabled"
    " <i>YASAPI/NT</i> starts the timer once initialized with an appropriate"
    " period matching a playback cycle. Otherwise the timer ist re-started"
    " again with each playback cycle."
    ;
  static const char attDevicePeriod[]=
    "Associated with <i>WASAPI</i> mode. According"
    " to <i>Microsoft</i> \"[<i>Default</i>] specifies the [...] period"
    " for a shared-mode stream. [<i>Minimum</i>] specifies the [...]"
    " period for an exclusive-mode stream.\""
    ;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  static const char attProAudio[]=
    "Should at least be enabled with mode <i>Exclusive</i> and"
    " strategy <i>Pull</i>."
    ;
#endif // ]
  static const char attAlignmentDance[]=
    "Do the \"alignment dance\" when error AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED"
    " occurs with IAudioClient::Initialize()."
    ;

  static ClassPage wasapi_rows_frame={
    { LABEL_STATIC(1,0,"Mode:",attMode),LABEL_MAPPED_DEVICE(2,0,
        wasapi.mode,eMode,mode_map,attMode) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.auto_convert_pcm,bAutoConvertPCM,
            attAutoConvertPCM),
        LABEL_STATIC(1,0,"Auto Convert PCM",attAutoConvertPCM) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.src_default_quality,bSRCDefaultQuality,
            attSRCDefaultQuality),
        LABEL_STATIC(1,0,"SRC Default Quality",
            attSRCDefaultQuality) },
    { LABEL_STATIC(1,0,"Device Period:",attDevicePeriod),
        LABEL_MAPPED_DEVICE(2,0,wasapi.period,ePeriod,period_map,
            attDevicePeriod) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy),
        LABEL_MAPPED_DEVICE(2,0,wasapi.strategy,eStrategy,strategy_map,
            attStrategy) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.timer_periodic,bTimerPeriodic,
            attTimerPeriodic),
        LABEL_STATIC(1,0,"Periodic Timer",attTimerPeriodic) },
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    { LABEL_STATIC(1,0,"Associate with \"Pro Audio\" Task:",
        attProAudio), CHECK_BUTTON_DEVICE(2,0,wasapi.pro_audio,bProAudio,
            attProAudio) },
#endif // ]
    { LABEL_STATIC(1,0,"Alignment Dance:",attAlignmentDance),
        CHECK_BUTTON_DEVICE(2,0,wasapi.alignment_dance,bAlignmentDance,
            attAlignmentDance) },
    { { NULL } }
  };

  static ClassPage wasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,wasapi.device) },
    { FRAME(wasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap bitdepth_map[]={
    { "As Delivered by Winamp", "<i>As Delivered by Winamp</i>",
        eBitdepthDisabled },
    { "16 Bit", "<i>16 Bit</i>", eBitdepth16 },
    { "24 Bit", "<i>24 Bit</i>", eBitdepth24 },
    { "32 Bit", "<i>32 Bit</i>", eBitdepth32 },
    { NULL, NULL, -1 }
  };

  static ConfigScaleParameter ring_size_param={
    .min=2.0,
    .max=30.0,
    .step=0.1,
  };

  static ConfigScaleParameter latency_param={
    .min=0.0,
    .max=1000.0,
    .step=10.0,
  };

#if defined (YASAPI_PLAYBACK_BEGIN_MIN_NUM_CHUNKS) // [
  static ConfigScaleParameter playback_begin_param={
    .nRefOffs=offsetof(Dialog,config.yasapi.ring_size),
    .bRef=1,
  };
#else // ] [
  static ConfigScaleParameter playback_begin_param={
    .min=1.0,
    .max=10.0,
    .step=0.1,
  };
#endif // ]

  static const char attBitdepth[]=
    "Whether <i>YASAPI/NT</i> should convert to the choosen format."
    ;
  static const char attRingSize[]=
    "The size of <i>YASAPI/NT</i>'s buffer in terms of a chunks's size"
    " as delivered by <i>Winamp</i> or the current input plug-in,"
    " respectively."
    ;
  static const char attChunkSize[]=
    "The size of an audio packet as delivered by <i>Winamp</i> or the"
    " current input plug-in, respectively, in order that <i>YASAPI/NT</i>"
    " sends it to <i>WASAPI</i>."
    ;
  static const char attPlaybackBegin[]=
#if defined (YASAPI_PLAYBACK_BEGIN_MIN_NUM_CHUNKS) // [
    "The minimum number of audio chunks <i>Winamp</i> or the current"
    " input plugin-in, respectively, should send before <i>YASAPI/NT</i>"
    " starts to output, i.e. being cached by <i>YASAPI/NT</i> (cf."
    " <i>Max. Latency</i>)."
#else // ] [
    "The amount of sound which should be cached by <i>YASAPI/NT</i>"
    " before starting to output."
#endif // ]
    ;
  static const char attMaxLatency[]=
    "When <i>Winamp</i> switches to a new track it requests the output"
    " plug-in to report the maximum latency. <i>YASAPI/NT</i> is"
    " not able to do so before knowing the size of audio chunks"
    " <i>Winamp</i> or the current input plug-in, respectively, is"
    " going to send. This option is meant to resolve the dilemma: Have"
    " an eye at <i>Configure → Monitor</i> and after having run all your"
    " DSP/input plug-ins set this option accordingly."
    ;

  static ClassPage yasapi_rows_frame={
    { LABEL_STATIC(1,0,"Bit Depth:",attBitdepth),
        COMBO_BOX_DEVICE(1,0,yasapi.bitdepth,eBitdepth,bitdepth_map,
            attBitdepth) },
    { LABEL_STATIC(1,0,"Ring Buffer Size:",attRingSize),
        SCALE_DEVICE(2,0,yasapi.ring_size,xRingSize,&ring_size_param,
            attRingSize),
        LABEL_STATIC(3,0,"× Chunk Size",attChunkSize) },
#if defined (YASAPI_PLAYBACK_BEGIN_MIN_NUM_CHUNKS) // [
    { LABEL_STATIC(1,0,"Playback Begin:",attPlaybackBegin),
        SCALE_DEVICE(2,0,yasapi.playback_begin,xPlaybackBegin,
            &playback_begin_param,attPlaybackBegin),
        LABEL_STATIC(1,0,"×",NULL),
        LABEL_MAPPED_DEVICE(1,0,yasapi.period_playback_begin,ePeriod,
            period_map,attDevicePeriod),
        LABEL_STATIC(1,0,"Buffer Size",NULL) },
#endif // ]
    { LABEL_STATIC(1,0,"Max. Latency:",attMaxLatency),
        SCALE_DEVICE(2,0,yasapi.max_latency_assumed,xMaxLatencyAssumed,
            &latency_param,attMaxLatency),
            LABEL_STATIC(3,0,"ms (Educated Guess)",NULL) },
#if ! defined (YASAPI_PLAYBACK_BEGIN_MIN_NUM_CHUNKS) // [
    { LABEL_STATIC(1,0,"Playback Begin:",attPlaybackBegin),
        SCALE_DEVICE(2,0,yasapi.playback_begin,xPlaybackBegin,
            &playback_begin_param,attPlaybackBegin),
        LABEL_STATIC(1,0,"×",NULL),
        LABEL_MAPPED_DEVICE(1,0,yasapi.period_playback_begin,ePeriod,
            period_map,attDevicePeriod),
        LABEL_STATIC(1,0,"Buffer Size",NULL) },
#endif // ]
    { { NULL } }
  };

  static ClassPage yasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,yasapi.device) },
    { FRAME(yasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigLevelBarParameter param_monitor_written={ 0 };
  static ConfigLevelBarParameter param_monitor_padding={ 0 };

  static const char attWinampNumTracks[]=
    "The overall number of individual tracks played with the current"
    " <i>Winamp</i> session and any device."
    ;
  static const char attWinampNumGaplessSequences[]=
    "The overall number of gapless track sequences played with the current"
    " <i>Winamp</i> session and any device. A gapless track sequence may"
    " in turn consist of several individual tracks sharing the same format."
    " Tracks of different format cannot played gaplessly"
    ;
  static const char attWinampDropouts[]=
    "The overall number of dropouts with the current <i>Winamp</i> session"
    " and any device."
    ;
  static const char attWinampMaxLatency[]=
    "The latency accumulated so far with the current <i>Winamp</i> session"
    " and any device."
    ;
  static const char attPadding[]=
    "Taken at the beginning of a playback cycle, the amount of not yet"
    " played sound cached in the <i>WASAPI</i> device buffer."
    " (Not defined in case of double buffering, i.e. exclusive pulled"
    " playback. Remains empty in this case.)"
    ;
  static const char attWritten[]=
    "Taken at the beginning of a playback cycle, the amount of not yet"
    " played sound cached in <i>YASAPI/NT</i>'s buffer."
    ;
  static const char attCurDevice[]=
    "The device in use."
    ;
  static const char attTracksEngine[]=
    "The number of tracks played within the current gapless sequence."
    ;
  static const char attFramesDevice[]=
    "The size of the buffer shared with the <i>WASAPI</i> device."
    ;
  static const char attFramesYasapi[]=
    "The size of <i>YASAPI/NT</i>'s buffer."
    ;
  static const char attEngineDropouts[]=
    "The number of dropouts within the current gapless sequence."
    ;
  static const char attLatencyGapless[]=
    "The latency accumulated so far with the current gapless track sequence."
    ;

  static ClassPage monitor_rows_frame={
    // {
    { LABEL_STATIC(1,0,"<b>Winamp Session</b>",NULL) },
    { LABEL_STATIC(1,0,"#Tracks:",attWinampNumTracks),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_tracks,stats.uNumTracks,attWinampNumTracks) },
    { LABEL_STATIC(1,0,"#Gapless Sequences:",attWinampNumGaplessSequences),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_gapless,stats.uNumTracks,
            attWinampNumGaplessSequences) },
    { LABEL_STATIC(1,0,"#Dropouts:",attWinampDropouts),
        LABEL_UINT32_PLAYER(1,0,"%u",monitor.player.num_dropouts,
            stats.uNumDropouts,attWinampDropouts) },
    { LABEL_STATIC(1,0,"Max. Latency:",attWinampMaxLatency),
        LABEL_UINT32_PLAYER(1,0,
        "%u ms",monitor.player.ms_latency,stats.uMsLatencyMax,
            attWinampMaxLatency) },
    // }
    { LABEL_STATIC(1,0,"",NULL) },
    // {
    { LABEL_STATIC(1,0,"<b>Current Track"
        " / Current Gapless Track Sequence</b>",NULL) },
    { LABEL_STATIC(1,0,"Device:",attCurDevice),
        LABEL_ENGINE(1,0,monitor.device,params.pFriendlyName,attCurDevice) },
#if defined (YASAPI_OPTION_VOLUME) // [
    { LABEL_STATIC(1,0,"Volume Control:",NULL),
        LABEL_VOLUME(1,0,monitor.engine.volume,NULL) },
#endif // ]
    { LABEL_STATIC(1,0,"#Tracks:",attTracksEngine), LABEL_UINT32_ENGINE(1,0,
        "%u",monitor.engine.num_tracks,stats.uNumTracks,attTracksEngine) },
    // {
    { LABEL_STATIC(1,0,"WASAPI Device Buffer",NULL) },
    { LABEL_STATIC(1,0,"Size:",attFramesDevice), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_device,params.uNumFramesDevice,
            attFramesDevice) },
    { LABEL_STATIC(1,0,"Padding:",attPadding), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_padding,monitor.padding,monitor.xPadding,attPadding) },
    // }
    // {
    { LABEL_STATIC(1,0,"YASAPI/NT Buffer",NULL) },
    { LABEL_STATIC(1,0,"#Dropouts:",attEngineDropouts),
        LABEL_UINT32_ENGINE(1,0,"%u",monitor.engine.num_dropouts,
            stats.uNumDropouts,attEngineDropouts) },
    { LABEL_STATIC(1,0,"Size:",attFramesYasapi), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_ring,params.uNumFramesRing,
            attFramesYasapi) },
    { LABEL_STATIC(1,0,"Max. Latency:",attLatencyGapless),
        LABEL_FRAMES_ENGINE(1,0,monitor.engine.num_frames_latency,
            stats.uNumFramesMax,attLatencyGapless) },
    { LABEL_STATIC(1,0,"Padding:",attWritten), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_written,monitor.written,monitor.xWritten,attWritten) },
    // }
    // }
    { { NULL } }
  };

  static ClassPage monitor_rows={
    { FRAME(monitor_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ClassPage explore_rows_frame={
    { { NULL },
        LABEL_MATRIX_BITS(1,1,0),
        LABEL_MATRIX_BITS(1,1,1),
        LABEL_MATRIX_BITS(1,1,2) },
    { LABEL_MATRIX_FREQ(1,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,2) },
    { LABEL_MATRIX_FREQ(1,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,2) },
    { LABEL_MATRIX_FREQ(1,1,2),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,2) },
    { LABEL_MATRIX_FREQ(1,1,3),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,2) },
    { LABEL_MATRIX_FREQ(1,1,4),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,2) },
    { LABEL_MATRIX_FREQ(1,1,5),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,2) },
    { LABEL_MATRIX_FREQ(1,1,6),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,2) },
    { LABEL_MATRIX_FREQ(1,1,7),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,2) },
    { { NULL } }
  };

  static ClassPage explore_rows={
    { LABEL_DEVICE_WASAPI(1,0,explore.device) },
    { LABEL_STATIC(1,0,"Matrix of <i>WASAPI</i> Supported Integer Formats"
        " for Stereo (2.0)/Exclusive",NULL) },
    { FRAME(explore_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ClassNotebook notebook_pages={
    { LABEL_STATIC(1,0,"Common","Configure options independent of the"
        " chosen device."), PAGE(common_rows) },
    { LABEL_STATIC(1,0,"WASAPI","Configure options specific to the"
        " chosen device w.r.t <i>Microsoft WASAPI</i>."), PAGE(wasapi_rows) },
    { LABEL_STATIC(1,0,"YASAPI/NT","Configure options specific to the"
        " chosen device w.r.t the <i>YASAPI/NT</i> plug-in."),
            PAGE(yasapi_rows) },
    { LABEL_STATIC(1,0,"Monitor","Monitor running <i>YASAPI/NT</i>. Nothing"
        " to configure."), PAGE(monitor_rows) },
    { LABEL_STATIC(1,0,"Explore","Explore capabilities of <i>Microsoft</i>"
        " <i>WASAPI</i> w.r.t the chosen device. Nothing to configure."),
        PAGE(explore_rows) },
    { { NULL } }
  };

  ////////////////
  static ClassPage rows={
    { PAGE(general_rows) },
    { NOTEBOOK(notebook_pages) },
    { BUTTON_BOX() },
    { { NULL } }
  };

  static ClassCell cell=PAGE(rows);

  /////////////////////////////////////////////////////////////////////////////
  GtkWidget *window,*widget;
  Expand e;

  // window.
  pDialog->window = window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  // box.
  e.h = FALSE;
  e.v = FALSE;
  widget = cell.lpVtbl->create (&cell,pDialog,&e);
  g_object_set (G_OBJECT (widget), "margin", ePanelSpace, NULL);

  if (e.h) {
    gtk_widget_set_hexpand (widget, TRUE);
    gtk_widget_set_halign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_halign (widget, GTK_ALIGN_CENTER);

  if (e.v) {
    gtk_widget_set_vexpand (widget, TRUE);
    gtk_widget_set_valign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_valign (widget, GTK_ALIGN_CENTER);

  gtk_container_add (GTK_CONTAINER (window), widget);
  DWRITELN("\"box\" added to \"window\".");

  // It's better to call "set()" on all of our widgets after all of them
  // have been initialized because some of them may have side effects on
  // others.
  pDialog->lpVtbl->set(pDialog);
}

static void ConfigApply(Dialog *pDialog, int moffs, const char *op)
{
#if 0 // [
  enum { eNumWidgets=&pDialog->config.explore.device
      - &pDialog->config.general.device
      + 1 };
#else // ] [
  enum {
    eWidgetSize=sizeof (ConfigWidget),
    eNumWidgetsGeneral=sizeof (pDialog->config.general)/eWidgetSize,
    eNumWidgetsCommon=sizeof (pDialog->config.common)/eWidgetSize,
    eNumWidgetsWasapi=sizeof (pDialog->config.wasapi)/eWidgetSize,
    eNumWidgetsYasapi=sizeof (pDialog->config.yasapi)/eWidgetSize,
    eNumWidgetsMonitor=sizeof (pDialog->config.monitor)/eWidgetSize,
    eNumWidgetsExplore=sizeof (pDialog->config.explore)/eWidgetSize,
    eNumWidgets
        =eNumWidgetsGeneral
        +eNumWidgetsCommon
        +eNumWidgetsWasapi
        +eNumWidgetsYasapi
        +eNumWidgetsMonitor
        +eNumWidgetsExplore,
  };
#endif // ]
  typedef void (method_t)(ConfigWidget *pWidget);

  ConfigWidget *pWidget=pDialog->config.widgets;
  ConfigWidget *pWidgetMax=pWidget + eNumWidgets;
  method_t *method;

//DVWRITELN("%s { ===================================================",op);
  while (pWidget<pWidgetMax) {
    if (pWidget->lpVtbl) {
      DWRITELN (pWidget->pCell->pId);
      method=*(method_t **)((char *)pWidget->lpVtbl + moffs);
      method(pWidget);
    }
    else
      DVWRITELN ("*** missing lpVtbl: %d ***",pWidget-pDialog->config.widgets);

    ++pWidget;
  }
//DVWRITELN("%s } ===================================================",op);
}

static void ConfigSet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,set),"set");
}

static void ConfigGet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,get),"get");
}

IDialogVtbl gIConfigVtbl={
#if defined (YASAPI_DEBUG) // [
  .pIdA="Config",
#endif // ]
  .pIdW=L"Config",
  .pTitle="Configure " YASAPI_TITLE_U8,
  .nWidthPixels=eConfigWidthPixels,
  .WindowCreate=ConfigWindowCreate,
  .set=ConfigSet,
  .get=ConfigGet,
};
