/*
 * yasapi_config_config_button_box.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_CONFIG_BUTTON_PRINT
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
static void button_print (GtkWidget *widget, gpointer data)
{
  DVWRITELN("button \"%s\" hit",(char *)data);
}
#endif // ]

static void button_rollback (GtkWidget *widget, gpointer data)
{
  Dialog *pDialog = data;
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Wasapi *pWasapi=PlayerWasapi(pPlayer);

  DWRITELN("rolling back ...");
  /*
  pDialog->lpVtbl->get(pDialog);
  WasapiSync (pWasapi);
  */
  OptionsRollback (&pWasapi->options);
  pDialog->lpVtbl->set(pDialog);
}

static void button_commit (GtkWidget *widget, gpointer data)
{
  Dialog *pDialog = data;
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Wasapi *pWasapi=PlayerWasapi(pPlayer);

  DWRITELN("saving ...");
  pDialog->lpVtbl->get(pDialog);
  WasapiSync (pWasapi);
  pDialog->lpVtbl->set(pDialog);
  OptionsSave (&pWasapi->options);
}

/*
static void button_restore (GtkWidget *widget, gpointer data)
{
  Dialog *pDialog = data;
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);

  DWRITELN("restoring ...");
  WasapiRestore (pWasapi);
  pDialog->lpVtbl->set(pDialog);
}
*/

/*
static void button_cancel (GtkWidget *widget, gpointer data)
{
  Dialog *pDialog = data;
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);

  DWRITELN("canceling ...");
  OptionsRollback (&pWasapi->options);
  pDialog->lpVtbl->set(pDialog);
}
*/

static void button_default (GtkWidget *widget, gpointer data)
{
  Dialog *pDialog = data;
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);

  DWRITELN("setting to default ...");
  WasapiDefault (pWasapi);
  pDialog->lpVtbl->set(pDialog);
}

static GtkWidget *ClassButtonBoxCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  GtkWidget *hbox,*button_box,*button;
  char *label;

  /////////////////////////////////////////////////////////////////////////////
  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL,0);
  DVWRITELN("\"hbox\" created: %p.",hbox);

  /////////////////////////////////////////////////////////////////////////////
  // button box
  button_box = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
  DVWRITELN("\"button_box\" created: %p.",button_box);
  gtk_box_pack_start (GTK_BOX (hbox),button_box,0,0,0);
  DWRITELN("\"button_box\" added to \"hbox\".");

  // button "Default".
  label = "Default";
  button = gtk_button_new_with_label (label);
  DVWRITELN("\"button\" created: %p.",button);
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_print), label);
#endif // ]
//#if 0 // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_default), pDialog);
//#endif // ]
  gtk_container_add (GTK_CONTAINER (button_box), button);
  DWRITELN("\"button\" added to \"button_box\".");
  gtk_widget_set_tooltip_text(button, "Refresh with default values. Please"
      " note that all devices are affected and not just the visible one.");

  // button box.
  button_box = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
  DVWRITELN("\"button_box\" created: %p.",button_box);
  gtk_box_pack_end (GTK_BOX (hbox),button_box,0,0,0);
  DWRITELN("\"button_box\" added to \"hbox\".");

  // button "Commit".
  label = "Commit";
  button = gtk_button_new_with_label (label);
  DVWRITELN("\"button\" created: %p.",button);
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_print), label);
#endif // ]
  g_signal_connect (button, "clicked", G_CALLBACK (button_commit), pDialog);
  gtk_container_add (GTK_CONTAINER (button_box), button);
  DWRITELN("\"button\" added to \"button_box\".");
  gtk_widget_set_tooltip_text(button, "Synchronize plug-in and disk.");

  // button "Rollback".
  label = "Rollback";
  button = gtk_button_new_with_label (label);
  DVWRITELN("\"button\" created: %p.",button);
  gtk_widget_set_margin_start (button, ePanelSpace);
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_print), label);
#endif // ]
  g_signal_connect (button, "clicked", G_CALLBACK (button_rollback), pDialog);
  gtk_container_add (GTK_CONTAINER (button_box), button);
  DWRITELN("\"button\" added to \"button_box\".");
  gtk_widget_set_tooltip_text(button, "Refresh with the latest synchronized"
      " state. Please note that all devices are affected and not just the"
      " visible one.");

  // button "Ok".
  label = "Ok";
  button = gtk_button_new_with_label (label);
  DVWRITELN("\"button\" created: %p.",button);
  gtk_widget_set_margin_start (button, ePanelSpace);
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_print), label);
#endif // ]
  g_signal_connect (button, "clicked", G_CALLBACK (button_commit), pDialog);
  g_signal_connect (button, "clicked", G_CALLBACK (gtk_dialog_options_save),
      pDialog);
  g_signal_connect_swapped (button, "clicked",
      G_CALLBACK (gtk_widget_hide), pDialog->window);
  gtk_container_add (GTK_CONTAINER (button_box), button);
  DWRITELN("\"button\" added to \"button_box\".");
  gtk_widget_set_tooltip_text(button, "Hide the window and commit.");

  // button "Cancel".
  label = "Cancel";
  button = gtk_button_new_with_label (label);
  DVWRITELN("\"button\" created: %p.",button);
  gtk_widget_set_margin_start (button, ePanelSpace);
#if defined (YASAPI_CONFIG_BUTTON_PRINT) // [
  g_signal_connect (button, "clicked", G_CALLBACK (button_print), label);
#endif // ]
  //g_signal_connect (button, "clicked", G_CALLBACK (button_cancel), pDialog);
  //g_signal_connect (button, "clicked", G_CALLBACK (button_restore), pDialog);
  g_signal_connect (button, "clicked", G_CALLBACK (button_rollback), pDialog);
  g_signal_connect (button, "clicked", G_CALLBACK (gtk_dialog_options_save),
      pDialog);
  g_signal_connect_swapped (button, "clicked",
      G_CALLBACK (gtk_widget_hide), pDialog->window);
  gtk_container_add (GTK_CONTAINER (button_box), button);
  DWRITELN("\"button\" added to \"button_box\".");
  gtk_widget_set_tooltip_text(button, "Hide the window and rollback.");

  pExpand->h = TRUE;

  return hbox;
}

IClassCellVtbl gcIClassButtonBoxVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="ButtonBox",
#endif // ]
  .create=ClassButtonBoxCreate,
};
