/*
 * yasapi_config_check_button.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>
#include <yasapi.h>

static void CheckButtonSet(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  int *pbActive=(int *)pWidget->pCell->nav(pDialog,pWidget->pCell->offs);
  GtkWidget *widget = pWidget->widget;
  OptionsMemberDevice *pMember
      =pDialog->pGtk->pPlayer->wasapi.selected.pDeviceConfig->pMember;
  ConfigWidget *pWidgetAutoConvertPCM
      =&pDialog->pGtk->config.config.wasapi.auto_convert_pcm;
  int bSensitive;

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), *pbActive);

  if (pWidget!=&pDialog->pGtk->config.config.wasapi.src_default_quality)
    goto exit;

  bSensitive=pMember->config.bAutoConvertPCM
      &&gtk_widget_get_sensitive (pWidgetAutoConvertPCM->widget);
  gtk_widget_set_sensitive (pWidget->widget, bSensitive);
exit:
  return;
}

static void CheckButtonGet(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  GtkWidget *widget = pWidget->widget;
  int bEnabled = gtk_widget_get_sensitive (widget);
  int *pbActive;

  if (bEnabled) {
    pbActive=(int *)pWidget->pCell->nav(pDialog,pWidget->pCell->offs);
    *pbActive = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
  }
}

static void yasapi_config_check_button_toggled (GtkComboBox *widget,
    gpointer user_data)
{
  Dialog *pDialog = user_data;
  ConfigWidget *pWidget = (ConfigWidget *)g_object_get_data(G_OBJECT(widget),
      CONFIG_WIDGET);

  pWidget->lpVtbl->get(pWidget);
  pDialog->lpVtbl->set(pDialog);
}

///////////////////////////////////////////////////////////////////////////////
static GtkWidget *ClassCheckButtonCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
#if defined (YASAPI_DEBUG) // [
    .pId="CheckButton",
#endif // ]
    .set=CheckButtonSet,
    .get=CheckButtonGet,
  };

  GtkWidget *widget = gtk_check_button_new();

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  g_signal_connect (widget, "toggled",
      G_CALLBACK (yasapi_config_check_button_toggled), pDialog);

  return widget;
}

IClassCellVtbl gcIClassCheckButtonVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="CheckButton",
#endif // ]
  .create=ClassCheckButtonCreate,
};
