/*
 * yasapi_config_combo_box_device.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

static void ConfigComboBoxDeviceSet(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  wchar_t *pCommonId=pWasapi->options.common.config.id;
  UINT nDevice;
  WasapiDevice *pDevice;
  const wchar_t *pDeviceId;

  for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
    pDevice=pWasapi->pDevices+nDevice;
    pDeviceId=pDevice->lpVtbl->getid(pDevice);

    if (!wcscmp(pCommonId,pDeviceId)) {
      gtk_combo_box_set_active (GTK_COMBO_BOX (pWidget->widget), nDevice);
      return;
    }
  }
}

static void ConfigComboBoxDeviceGet(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  UINT nDevice=gtk_combo_box_get_active (GTK_COMBO_BOX (pWidget->widget));
  WasapiDevice *pDevice=pWasapi->pDevices+nDevice;
  const wchar_t *pDeviceId=pDevice->lpVtbl->getid(pDevice);
  wchar_t *pCommonId=pWasapi->options.common.config.id;

  wcscpy(pCommonId,pDeviceId);
  pWasapi->selected.pDeviceConfig=pDevice;

  DVWRITELNW(L"*** Got \"%s\" ***", pCommonId);
}

static void yasapi_config_device_combo_box_changed (GtkComboBox *widget,
    gpointer user_data)
{
  Dialog *pDialog = user_data;
  ConfigWidget *pWidget = (ConfigWidget *)g_object_get_data(G_OBJECT(widget),
      CONFIG_WIDGET);

  pWidget->lpVtbl->get(pWidget);
  pDialog->lpVtbl->set(pDialog);
}

static GtkWidget *ConfigComboBoxDeviceCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
#if defined (YASAPI_DEBUG) // [
    .pId="ComboBoxDevice",
#endif // ]
    .set=ConfigComboBoxDeviceSet,
    .get=ConfigComboBoxDeviceGet,
  };

  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  GtkWidget *widget = gtk_combo_box_text_new();
  UINT nDevice;

  for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
    gtk_combo_box_text_append (GTK_COMBO_BOX_TEXT (widget), NULL, 
        pWasapi->pDevices[nDevice].pFriendlyName);
  }

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  g_signal_connect (widget, "changed",
      G_CALLBACK (yasapi_config_device_combo_box_changed), pDialog);

  return widget;
}

IClassCellVtbl gcIClassComboBoxDeviceVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="ComboBoxDevice",
#endif // ]
  .create=ConfigComboBoxDeviceCreate,
};
