/*
 * yasapi_config_matrix.h
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

///////////////////////////////////////////////////////////////////////////////
static GtkWidget *ConfigLabelMatrixBitsCreate(ClassCell *pCell,
    Dialog *pDialog, Expand *pExpand)
{
  int col=(int)pCell->data;
  const char *label;
  GtkWidget *widget;

  label=gaBitDepths[col].label;
  widget = gtk_label_new (label);

  return widget;
}

IClassCellVtbl gcIClassLabelMatrixBitsVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="LabelMatrixBits",
#endif // ]
  .create=ConfigLabelMatrixBitsCreate,
};

///////////////////////////////////////////////////////////////////////////////
static GtkWidget *ConfigLabelMatrixFreqCreate(ClassCell *pCell,
    Dialog *pDialog, Expand *pExpand)
{

  int row=(int)pCell->data;
  const char *label;
  GtkWidget *widget;

  label=gaFrequencies[row].label;
  widget = gtk_label_new (label);

  return widget;
}

IClassCellVtbl gcIClassLabelMatrixFreqVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="LabelMatrixFreq",
#endif // ]
  .create=ConfigLabelMatrixFreqCreate,
};

///////////////////////////////////////////////////////////////////////////////
static void CheckButtonMatrixSet(ConfigWidget *pWidget)
{
  Wasapi *pWasapi=PlayerWasapi(pWidget->pDialog->pGtk->pPlayer);
  WasapiDevice *pDeviceConfig=pWasapi->selected.pDeviceConfig;
  OptionsMemberDevice *pMember=pDeviceConfig->pMember;
  FormatMatrix *pMatrix=&pMember->matrix;
  int *pbActive=(int *)((char *)pMatrix+pWidget->pCell->offs);
  GtkWidget *widget=pWidget->widget;

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), *pbActive);
}

static void CheckButtonMatrixGet(ConfigWidget *pWidget)
{
}

static GtkWidget *ConfigCheckButtonMatrixCreate(ClassCell *pCell,
    Dialog *pDialog, Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
#if defined (YASAPI_DEBUG) // [
    .pId="CheckButtonMatrix",
#endif // ]
    .set=CheckButtonMatrixSet,
    .get=CheckButtonMatrixGet,
  };

  GtkWidget *widget = gtk_check_button_new();

  gtk_widget_set_sensitive (widget, FALSE);
  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  return widget;
}

IClassCellVtbl gcIClassCheckButtonMatrixVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="CheckButtonMatrix",
#endif // ]
  .create=ConfigCheckButtonMatrixCreate,
};
