/*
 * yasapi_config_root.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>

static int ClassPageCountColumns(ClassCell *pCell)
{
  ClassPage *pPage = pCell->data;
  int cCols=0,nCol,nSpan;
  ClassRow *pRow;

  for (pRow=*pPage;(*pRow)[0].lpVtbl||(*pRow)[1].lpVtbl;++pRow) {
    pCell=*pRow;
    nCol=0;

    if (!pCell[1].lpVtbl)
      ++nCol;
    else {
      ++pCell;
      ++nCol;

      while (pCell->lpVtbl) {
        nSpan=0<pCell->nSpan?pCell->nSpan:1;
        ++pCell;
        nCol+=nSpan;
      }
    }

    if (cCols<nCol)
      cCols=nCol;
  }

DVWRITELN("*** cCols=%d ***",cCols);
  return cCols;
}

static GtkWidget *ClassPageCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  ClassPage *pPage = pCell->data;
  int cCols=ClassPageCountColumns(pCell);
  ClassRow *pRow;
  int nRow,nCol,nSpan;
  GtkWidget *grid,*widget;
  Expand e;

  grid = gtk_grid_new ();
  gtk_grid_set_row_spacing (GTK_GRID (grid), ePanelSpace);
  gtk_grid_set_column_spacing (GTK_GRID (grid), ePanelSpace);

  for (pRow=*pPage;(*pRow)[0].lpVtbl||(*pRow)[1].lpVtbl;++pRow) {
    pCell=*pRow;
    nRow=pRow-*pPage;
    nCol=0;

    if (!pCell[1].lpVtbl) {
      e.h = FALSE;
      e.v = FALSE;
      widget = pCell->lpVtbl->create(pCell,pDialog,&e);

      if (e.h) {
//fprintf (stderr, "*** %s\n",pCell->lpVtbl->pId); fflush (stderr);
        gtk_widget_set_hexpand (widget, TRUE);
        gtk_widget_set_halign (widget, GTK_ALIGN_FILL);
        pExpand->h=TRUE;
      }
      else
        gtk_widget_set_halign (widget, GTK_ALIGN_CENTER);

      if (e.v) {
        gtk_widget_set_vexpand (widget, TRUE);
        gtk_widget_set_valign (widget, GTK_ALIGN_FILL);
        pExpand->v=TRUE;
      }

      gtk_grid_attach (GTK_GRID (grid), widget, nCol, nRow, cCols, 1);
    }
    else {
      if (pCell->lpVtbl) {
        widget = pCell->lpVtbl->create(pCell,pDialog,&e);

        gtk_widget_set_halign (widget, GTK_ALIGN_END);
        gtk_grid_attach (GTK_GRID (grid), widget, nCol, nRow, 1, 1);
      }

      ++pCell;
      ++nCol;

      while (nCol<eClassRowNumCols&&pCell->lpVtbl) {
        e.h = FALSE;
        e.v = FALSE;
        widget = pCell->lpVtbl->create(pCell,pDialog,&e);
        nSpan=0<pCell->nSpan?pCell->nSpan:1;

        if (e.h) {
          gtk_widget_set_hexpand (widget, TRUE);
          pExpand->h=e.h;
        }
        else {
          if ((nCol+nSpan==eClassRowNumCols||!pCell[1].lpVtbl)&&!pExpand->h) {
            // This row's final column and none of the this row's colums have
            // requested hexpand: all this row's additional space needs to
            // go here.
            gtk_widget_set_hexpand (widget, TRUE);
          }

          if (pCell->bCenter)
            gtk_widget_set_halign (widget, GTK_ALIGN_CENTER);
          else
            gtk_widget_set_halign (widget, GTK_ALIGN_START);
        }

        gtk_grid_attach (GTK_GRID (grid), widget, nCol, nRow, nSpan, 1);

        ++pCell;
        nCol+=nSpan;
      }
    }
  }

  return grid;
}

IClassCellVtbl gcIClassPageVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="Page",
#endif // ]
  .create=ClassPageCreate,
};

static GtkWidget *ClassFrameCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  GtkWidget *grid,*widget;
  Expand e;

  grid = gtk_grid_new ();
  e.h=FALSE;
  e.v=FALSE;
#if 1 // [
  widget=ClassPageCreate(pCell,pDialog,&e);
#else // ] [
  pCell = pCell->data;
  widget=pCell->lpVtbl->create(pCell,pDialog,&e);
#endif // ]
  gtk_widget_set_hexpand (widget, TRUE);
  gtk_widget_set_halign (widget, e.h?GTK_ALIGN_FILL:GTK_ALIGN_CENTER);
  gtk_widget_set_vexpand (widget, TRUE);
  gtk_widget_set_valign (widget, GTK_ALIGN_CENTER);
  gtk_grid_attach (GTK_GRID (grid), widget, 1, 1, 1, 1);
  pExpand->h = TRUE;
  pExpand->v = TRUE;

  return grid;
}

IClassCellVtbl gcIClassFrameVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="Frame",
#endif // ]
  .create=ClassFrameCreate,
};
