/*
 * yasapi_config_progess_bar.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>

static void ConfigLevelBarSet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  double *pxValue=(double *)pCell->nav(pWidget->pDialog,pCell->offs);

  if (pxValue&&*pxValue) {
    gtk_progress_bar_set_fraction  (GTK_PROGRESS_BAR (pWidget->widget),
        *pxValue);
  }
  else
    gtk_progress_bar_set_fraction  (GTK_PROGRESS_BAR (pWidget->widget), 0.0);
}

static void ConfigLevelBarGet(ConfigWidget *pWidget)
{
}

static GtkWidget *ConfigLevelBarCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
#if defined (YASAPI_DEBUG) // [
    .pId="LevelBar",
#endif // ]
    .set=ConfigLevelBarSet,
    .get=ConfigLevelBarGet,
  };

  // For GTK 3.20 and newer, when defining the CSS, it seems that, instead
  // of using the progressbar class, you must use the progress and trough
  // classes. This is what worked for me:
  //
  // progress, trough {
  //    min-height: 30px;
  // }
  //
  // Gtk Progress bar height
  // https://stackoverflow.com/questions/44290651/gtk-progress-bar-height
  static gchar data[]=
    "progress, trough {"
    "\n   min-height: 24px;"
    "\n}"
    ;
  Gtk *pGtk=pDialog->pGtk;
  GtkWidget *widget;
  GtkStyleContext *context;

  if (!pGtk->provider) {
    pGtk->provider = gtk_css_provider_new();
    gtk_css_provider_load_from_data (pGtk->provider, data, -1, NULL);
  }

  widget = gtk_progress_bar_new ();
  context = gtk_widget_get_style_context(widget);

  gtk_style_context_add_provider (context,
      GTK_STYLE_PROVIDER (pGtk->provider), GTK_STYLE_PROVIDER_PRIORITY_USER);

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);
  pExpand->h=TRUE;

  return widget;
}

IClassCellVtbl gcIClassLevelBarVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="LevelBar",
#endif // ]
  .create=ConfigLevelBarCreate,
};
