/*
 * yasapi_config_scale.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>

static void ConfigScaleSet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  double *value=(double *)pCell->nav(pWidget->pDialog,pCell->offs);

  gtk_range_set_value (GTK_RANGE (pWidget->widget), *value);
}

static void ConfigScaleGet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  double *value=(double *)pCell->nav(pWidget->pDialog,pCell->offs);

  *value = gtk_range_get_value (GTK_RANGE (pWidget->widget));
}

static GtkWidget *ConfigScaleCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
#if defined (YASAPI_DEBUG) // [
    .pId="Scale",
#endif // ]
    .set=ConfigScaleSet,
    .get=ConfigScaleGet,
  };

  ConfigScaleParameter *p=pCell->data;
  GtkWidget *widget;
#if defined (YASAPI_PLAYBACK_BEGIN_MIN_NUM_CHUNKS) // [
  ConfigWidget *ref;

  while (p->bRef) {
    ref = (ConfigWidget *)((char *)pDialog+p->nRefOffs);
    p = ref->pCell->data;
  }
#endif // ]

  widget = gtk_scale_new_with_range (GTK_ORIENTATION_HORIZONTAL, p->min,
      p->max, p->step);

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);
  pExpand->h= TRUE;

  return widget;
}

IClassCellVtbl gcIClassScaleVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="Scale",
#endif // ]
  .create=ConfigScaleCreate,
};
