/*
 * yasapi_gtk.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#if defined (YASAPI_FILE_OPEN_DIALOG) && defined (__GNUC__) // [
#include <yasapi_mingw32.h>
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_GTK_LOAD

///////////////////////////////////////////////////////////////////////////////
typedef enum _EGtkThread EGtkThread;
#if defined (YASAPI_GTK_LOAD) // [
typedef enum _EGtkThreadLimits EGtkThreadLimits;
#endif // ]
typedef struct _GtkThread GtkThread;
typedef struct _GtkLib GtkLib;

///////////////////////////////////////////////////////////////////////////////
enum _EGtkThread {
  eGtkThreadNull,
#if defined (YASAPI_GTK_LOAD) // [
  eGtkThreadGtkLoaded,
#endif // ]
  eGtkThreadDialogs,
  ////////////////
  eGtkThreadMax,
};

static DWORD WINAPI GtkThreadProc(LPVOID lpParameter);
#if defined (YASAPI_DEBUG) // [
static const char *GtkThreadState(const GtkThread *pThread);
#endif // ]

struct _GtkLib {
  const char *name;
  const gchar *(*check) (guint required_major, guint required_minor,
      guint required_micro);
  guint major;
  guint minor;
  guint micro;
  int bLibGlib;
};

///////////////////////////////////////////////////////////////////////////////
static int GtkCreateBase(Gtk *pGtk, Player *pPlayer)
{
  int code=-1;

  /////////////////////////////////////////////////////////////////////////////
  pGtk->eState=eGtkNull;
  pGtk->pPlayer=pPlayer;
  pGtk->pDialog=NULL;

  /////////////////////////////////////////////////////////////////////////////
  if (!pPlayer->ini[0])
    PlayerPropertiesPath(pPlayer);

  PropertiesRead(gcaPropertiesGtk,pPlayer->ini,gcaKeyGtk,
      &pGtk->options.original,&gcOptionsGtk);
  pGtk->options.current=pGtk->options.original;
  pGtk->options.config=pGtk->options.current;
  ++pGtk->eState;
  DASSERT(eGtkOptions==pGtk->eState,exit);
  DWRITELN(GtkState(pGtk));

  /////////////////////////////////////////////////////////////////////////////
  InitializeCriticalSection(&pGtk->cs);
  ++pGtk->eState;
  DASSERT(eGtkCriticalSection==pGtk->eState,exit);
  DWRITELN(GtkState(pGtk));

  /////////////////////////////////////////////////////////////////////////////
  if (!(pGtk->hProducer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating producer event");
    goto exit;
  }

  ++pGtk->eState;
  DASSERT(eGtkProducer==pGtk->eState,exit);
  DWRITELN(GtkState(pGtk));

  /////////////////////////////////////////////////////////////////////////////
  if (!(pGtk->hConsumer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pGtk->eState;
  DASSERT(eGtkConsumer==pGtk->eState,exit);
  DWRITELN(GtkState(pGtk));

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static void GtkDialogActivate (Dialog *pDialog)
{
  HWND hWnd=GDK_WINDOW_HWND(gtk_widget_get_window (pDialog->window));
  Gtk *pGtk=pDialog->pGtk;
  EDialogPos eDialogPos;
  Wasapi *pWasapi;
  HWND hWndParent;

  if (!(pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer))) {
    DMESSAGE("accessing wasapi");
    goto exit;
  }

  if (pDialog==&pGtk->config)
    eDialogPos=pWasapi->options.common.current.pos.eConfig;
  else if (pDialog==&pGtk->about)
    eDialogPos=pWasapi->options.common.current.pos.eAbout;
  else
    eDialogPos=eDialogPosFreeFloating;

  switch (eDialogPos) {
  case eDialogPosOnTopPreferences:
    hWndParent=pDialog->hWndParent;
    pDialog->hWndParent=NULL;
    SetWindowLongPtrW(hWnd,GWL_HWNDPARENT,(LONG_PTR)hWndParent);

    SetWindowPos(
      hWnd,                     // _In_     HWND hWnd,
      HWND_NOTOPMOST,           // _In_opt_ HWND hWndInsertAfter,
      0,                        // _In_     int  X,
      0,                        // _In_     int  Y,
      0,                        // _In_     int  cx,
      0,                        // _In_     int  cy,
      SWP_NOMOVE|SWP_NOSIZE     // _In_     UINT uFlags
    );

    break;
  case eDialogPosAlwaysOnTop:
    hWndParent=pDialog->hWndParent;
    pDialog->hWndParent=NULL;
    SetWindowLongPtrW(hWnd,GWL_HWNDPARENT,(LONG_PTR)hWndParent);

    SetWindowPos(
      hWnd,                     // _In_     HWND hWnd,
      HWND_TOPMOST,             // _In_opt_ HWND hWndInsertAfter,
      0,                        // _In_     int  X,
      0,                        // _In_     int  Y,
      0,                        // _In_     int  cx,
      0,                        // _In_     int  cy,
      SWP_NOMOVE|SWP_NOSIZE     // _In_     UINT uFlags
    );

    break;
  case eDialogPosFreeFloating:
  default:
    SetWindowLongPtrW(hWnd,GWL_HWNDPARENT,(LONG_PTR)NULL);
    break;
  }
exit:
  BringWindowToTop(hWnd);
  SetActiveWindow(hWnd);
}

static void GtkDialogShowSync (Dialog *pDialog, int bActivate)
{
  if (!pDialog||!pDialog->window)
    goto exit;

  if (!gtk_widget_get_visible (pDialog->window))
    gtk_widget_show_all (pDialog->window);

  if (bActivate) {
    if (!pDialog->bMoved) {
      if (0<=pDialog->options.nPosX&&0<=pDialog->options.nPosY) {
        gtk_window_move (GTK_WINDOW(pDialog->window),pDialog->options.nPosX,
            pDialog->options.nPosY);
      }

      if (pDialog->notebook&&0<=pDialog->options.nPage) {
        gtk_notebook_set_current_page (GTK_NOTEBOOK (pDialog->notebook),
            pDialog->options.nPage);
      }

      pDialog->bMoved=1;
    }

    GtkDialogActivate (pDialog);
    pDialog->pGtk->pDialog=NULL;
  }
exit:
  return;
}

static gboolean GtkDialogShowAsync (gpointer user_data)
{
  GtkDialogShowSync (user_data,1);

  return FALSE;
}

int GtkShow(Gtk *pGtk, Player *pPlayer, Dialog *pDialog, HWND hWndParent)
{
  int code=-1;

  if (pGtk->eState<eGtkConsumer&&GtkCreateBase(pGtk,pPlayer)<0) {
    DMESSAGE("creating gtk base");
    goto ebase;
  }

  EnterCriticalSection(&pGtk->cs); // {

  if (pGtk->pDialog)
    goto exit;

  pGtk->pDialog=pDialog;
  pDialog->hWndParent=hWndParent;

  if (pDialog->window)
    g_idle_add (GtkDialogShowAsync, pDialog);
  else {
    pGtk->hThread=CreateThread(
      NULL,     // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
      0ul,      // _In_      SIZE_T                 dwStackSize,
      GtkThreadProc,
                // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
      pGtk,     // _In_opt_  LPVOID                 lpParameter,
      0ul,      // _In_      DWORD                  dwCreationFlags,
      NULL      // _Out_opt_ LPDWORD                lpThreadId
    );

    if (!pGtk->hThread) {
      DMESSAGE("creating win thread");
      goto exit;
    }

    ++pGtk->eState;
    DASSERT(eGtkThread==pGtk->eState,exit);
    DWRITELN(GtkState(pGtk));
    SetEvent(pGtk->hConsumer);

    while (pGtk->eState<eGtkThreadRunning) {
      LeaveCriticalSection(&pGtk->cs); // {
      WaitForSingleObject(pGtk->hProducer,INFINITE);
      EnterCriticalSection(&pGtk->cs); // }
    }
  }

  code=0;
exit:
  LeaveCriticalSection(&pGtk->cs); // }
ebase:
  return code;
}

void GtkDestroy(Gtk *pGtk)
{
  switch (pGtk->eState) {
  case eGtkLoaded:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkLoaded==pGtk->eState,exit);
    --pGtk->eState;
  case eGtkThreadRunning:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkThreadRunning==pGtk->eState,exit);
    gtk_main_quit ();
    --pGtk->eState;
  case eGtkThread:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkThread==pGtk->eState,exit);
    WaitForSingleObject(pGtk->hThread,INFINITE);
    --pGtk->eState;
    CloseHandle(pGtk->hThread);
    pGtk->hThread=NULL;
  case eGtkConsumer:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkConsumer==pGtk->eState,exit);
    --pGtk->eState;
    CloseHandle(pGtk->hConsumer);
    pGtk->hThread=NULL;
  case eGtkProducer:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkProducer==pGtk->eState,exit);
    --pGtk->eState;
    CloseHandle(pGtk->hProducer);
    pGtk->hProducer=NULL;
  case eGtkCriticalSection:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkCriticalSection==pGtk->eState,exit);
    --pGtk->eState;
  case eGtkOptions:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkOptions==pGtk->eState,exit);
    GtkPathSave(pGtk);
    --pGtk->eState;
  case eGtkNull:
    DWRITELN(GtkState(pGtk));
    DASSERT(eGtkNull==pGtk->eState,exit);
  default:
    break;
  }
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_GTK_LOAD) // [
static GtkLib gaGtkLibs[]={
  { "libglib-2.0-0.dll",glib_check_version,
      GLIB_MAJOR_VERSION,GLIB_MINOR_VERSION,GLIB_MICRO_VERSION,1 },
  { "libgobject-2.0-0.dll",NULL,0,0,0 },
  { "libgdk-3-0.dll",NULL,
      GDK_MAJOR_VERSION,GDK_MINOR_VERSION,GDK_MICRO_VERSION },
  { "libgtk-3-0.dll",gtk_check_version,
      GTK_MAJOR_VERSION,GTK_MINOR_VERSION,GTK_MICRO_VERSION },
};

enum _EGtkThreadLimits {
  eGtkThreadCountLibs=(sizeof gaGtkLibs)/(sizeof gaGtkLibs[0]),
};
#endif // ]

struct _GtkThread {
  EGtkThread eState;
  Gtk *pGtk;
  HWND hWndParent;
  WNDPROC lpPrevWndFunc;
#if defined (YASAPI_GTK_LOAD) // [
  HMODULE ahGtkLibs[eGtkThreadCountLibs];
#endif // ]
  int nGtkLibs;
};

///////////////////////////////////////////////////////////////////////////////
// We subclass the parent window in order to display a wait cursor.
static LRESULT CALLBACK GtkWndProc(HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  switch (uMsg) {
  case WM_SETCURSOR:
    SetCursor(LoadCursor(0,IDC_WAIT));
    return TRUE;
  default:
    break;
  }

  return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

static int GtkThreadSubclass(GtkThread *pThread)
{
  int code=-1;

  if (!(pThread->hWndParent=pThread->pGtk->pDialog->hWndParent))
    goto success;

  pThread->lpPrevWndFunc=(WNDPROC)SetWindowLong(
    pThread->hWndParent,            // _In_ HWND hWnd,
    GWL_WNDPROC,                    // _In_ int  nIndex,
    (LONG)GtkWndProc                // _In_ LONG dwNewLong
  );

  if (!pThread->lpPrevWndFunc) {
    DMESSAGE("subclassing");
    goto exit;
  }
success:
  code=0;
exit:
  return code;
}

static void GtkThreadUnsubclass(GtkThread *pThread)
{
  if (pThread->hWndParent&&pThread->lpPrevWndFunc) {
    SetWindowLong(
      pThread->hWndParent,          // _In_ HWND hWnd,
      GWL_WNDPROC,                  // _In_ int  nIndex,
      (LONG)pThread->lpPrevWndFunc  // _In_ LONG dwNewLong
    );
  }
}

#if defined (YASAPI_GTK_LOAD) // [
///////////////////////////////////////////////////////////////////////////////
static void GtkThreadUnload(GtkThread *pThread)
{
  while (0<pThread->nGtkLibs) {
    --pThread->nGtkLibs;
    FreeLibrary(pThread->ahGtkLibs[pThread->nGtkLibs]);
    pThread->ahGtkLibs[pThread->nGtkLibs]=NULL;
    DVWRITELN("\"%s\" unloaded.",gaGtkLibs[pThread->nGtkLibs].name);
  }
}

static int GtkThreadLoad(GtkThread *pThread, const wchar_t *pPath)
{
  static const wchar_t aPathName[]=L"yagtk-rt\\bin";

  wchar_t path[MAX_PATH],*pp;
  GtkLib *pGtkLib=gaGtkLibs;
  HMODULE *phGtkLib=pThread->ahGtkLibs;
  DWORD dwSize;
  const gchar *check;

  /////////////////////////////////////////////////////////////////////////////
  pThread->nGtkLibs=0;

  if (pPath||pThread->pGtk->pPlayer->pOut) {
    if (pPath&&*pPath) {
      pp=path;
      wcscpy(pp,pPath);
      pp+=wcslen(pp);
      wcscpy(pp,L"\\bin");
    }
    else if (pThread->pGtk->pPlayer->pOut) {
      /////////////////////////////////////////////////////////////////////////
      dwSize=GetModuleFileNameW(
        pThread->pGtk->pPlayer->pOut->hDllInstance,
                        // _In_opt_ HMODULE hModule,
        path,           // _Out_    LPTSTR  lpFilename,
        sizeof path     // _In_     DWORD   nSize
      );

      for (pp=path+dwSize;path<pp&&'\\'!=pp[-1];--pp)
        ;

      wcscpy(pp,aPathName);
    }
    else {
      DMESSAGE("Missing path to library");
      goto epath;
    }

    if (!SetDllDirectoryW(path)) {
      DVMESSAGE("extending library path with error %lu",GetLastError());
      goto epath;
    }

    DVWRITELNW(L"Library search path extended by \"%s\".",path);
  }

  /////////////////////////////////////////////////////////////////////////////
  while (pThread->nGtkLibs<eGtkThreadCountLibs) {
    if (!(*phGtkLib=LoadLibraryA(pGtkLib->name))) {
      DVMESSAGE("loading \"%s\"",pGtkLib->name);
      goto eload;
    }

    DVWRITELN("\"%s\" loaded.",pGtkLib->name);

    if (pGtkLib->bLibGlib)
      pThread->pGtk->hLibGlib=*phGtkLib;

    if (pGtkLib->check)
      check=pGtkLib->check(pGtkLib->major,pGtkLib->minor,pGtkLib->micro);
    else
      check=NULL;

    ++pThread->nGtkLibs;
    ++pGtkLib;
    ++phGtkLib;

    if (check) {
      DVMESSAGE("\"%s\" check failed: \"%s\"",pGtkLib[-1].name,check);
      goto echeck;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  gtk_init(NULL,NULL);
  DWRITELN("GTK initialized.");

  return 0;
echeck:
eload:
  GtkThreadUnload(pThread);
epath:
  return -1;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int GtkThreadCreate(GtkThread *pThread)
{
  int code=0;
  Gtk *pGtk=pThread->pGtk;
  const wchar_t *pPath=pGtk->options.current.path;
  Player *pPlayer=pGtk->pPlayer;
  Dialog *pDialogMax=pGtk->aDialog+eGtkNumDialogs;
  Dialog *pDialog;

  switch (pThread->eState) {
  case eGtkThreadNull:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadNull==pThread->eState,exit);
#if defined (YASAPI_GTK_LOAD) // [
    ++pThread->eState;
  case eGtkThreadGtkLoaded:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadGtkLoaded==pThread->eState,exit);

    if (GtkThreadLoad(pThread,pPath)<0) {
      code=1;
      goto exit;
    }
#endif // ]
    ++pThread->eState;
  case eGtkThreadDialogs:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadDialogs==pThread->eState,exit);

    if (!pPlayer->ini[0])
      PlayerPropertiesPath(pPlayer);

    pGtk->about.lpVtbl=&gIAboutVtbl;
    pGtk->about.pGtk=pGtk;

    pGtk->config.lpVtbl=&gIConfigVtbl;
    pGtk->config.pGtk=pGtk;

    for (pDialog=pGtk->aDialog;pDialog<pDialogMax;++pDialog) {
      DASSERT(pDialog->lpVtbl,exit);
      DASSERT(pDialog->pGtk,exit);
      pDialog->bMoved=0;
      PropertiesRead(gcaPropertiesDialog,pPlayer->ini,pDialog->lpVtbl->pIdW,
          &pDialog->options,&gcOptionsDialog);
    }
  default:
    break;
  }

  code=0;
#if defined (YASAPI_GTK_LOAD) || defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return code;
}

static void GtkDialogGetOptions(Dialog *pDialog, OptionsDialog *pOptions)
{
  GtkWidget *window = pDialog->window;
  GtkWidget *notebook = pDialog->notebook;

  if (gtk_widget_get_visible (window)) {
    pOptions->nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
    gtk_window_get_position (GTK_WINDOW (window), &pOptions->nPosX,
        &pOptions->nPosY);
  }
}

static void GtkDialogWrite(Dialog *pDialog)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;
  OptionsDialog options=pDialog->options;

  GtkDialogGetOptions(pDialog,&options);
  PropertiesWrite(gcaPropertiesDialog,pPlayer->ini,pDialog->lpVtbl->pIdW,
      &options,&pDialog->options);
  pDialog->options=options;
}

void gtk_dialog_options_save (GtkWidget *widget, gpointer data)
{
  GtkDialogWrite(data);
}

static void GtkThreadDestroy(GtkThread *pThread)
{
  Gtk *pGtk=pThread->pGtk;
  Dialog *pDialog;

  switch (pThread->eState) {
  case eGtkThreadDialogs:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadDialogs==pThread->eState,exit);
    --pThread->eState;

    for (pDialog=pGtk->aDialog+eGtkNumDialogs;pGtk->aDialog<pDialog;)
      GtkDialogWrite(--pDialog);
#if defined (YASAPI_GTK_LOAD) // [
  case eGtkThreadGtkLoaded:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadGtkLoaded==pThread->eState,exit);
    --pThread->eState;
    GtkThreadUnload(pThread);
#endif // ]
  case eGtkThreadNull:
    DWRITELN(GtkThreadState(pThread));
    DASSERT(eGtkThreadNull==pThread->eState,exit);
  default:
    break;
  }
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return;
}

static gboolean window_delete_event (GtkWidget *window, GdkEvent *event,
    gpointer user_data)
{
  Dialog *pDialog=user_data;

  GtkDialogWrite(pDialog);
  DVWRITELN("nPosX=%d nPosV=%d nPage=%d",pDialog->options.nPosX,
      pDialog->options.nPosY,pDialog->options.nPage);
  gtk_widget_hide (window);

  return TRUE;
}

static int GtkThreadShowPathDialog(GtkThread *pThread)
{
  static const wchar_t aTitle[]=L"Choose the Root of the GTK Runtime";

  int code=-1;
  wchar_t *path=pThread->pGtk->options.current.path;
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
  HWND hWndParent=pThread->pGtk->pDialog->hWndParent;
  IFileDialog *pFileDialog;
  IShellItem *pItem;
  PWSTR pFilePath;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  *path=0;

  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_FileOpenDialog,    // _In_  REFCLSID  rclsid,
    NULL,                     // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_INPROC_SERVER,     // _In_  DWORD     dwClsContext,
    &IID_IFileDialog,         // _In_  REFIID    riid,
    (void **)&pFileDialog     // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("CoCreateInstance",REGDB_E_CLASSNOTREG);
    DCMESSAGE("CoCreateInstance",CLASS_E_NOAGGREGATION);
    DCMESSAGE("CoCreateInstance",E_NOINTERFACE);
    DCMESSAGE("CoCreateInstance",E_POINTER);
    DDMESSAGE("CoCreateInstance",hr);
    }

    goto ecreate;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->SetTitle(pFileDialog,aTitle);

  if (FAILED(hr))
    goto error;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->SetOptions(pFileDialog,FOS_PICKFOLDERS);

  if (FAILED(hr))
    goto error;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->Show(pFileDialog,hWndParent);

  if (FAILED(hr))
    goto error;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->GetResult(pFileDialog,&pItem);

  if (FAILED(hr))
    goto error;

  /////////////////////////////////////////////////////////////////////////////
  hr=pItem->lpVtbl->GetDisplayName(pItem,SIGDN_FILESYSPATH,&pFilePath);

  if (FAILED(hr))
    goto error;

  wcscpy(path,pFilePath);

  /////////////////////////////////////////////////////////////////////////////
  code=0;
error:
  CoTaskMemFree(pFilePath);
  pItem->lpVtbl->Release(pItem);
  pFileDialog->lpVtbl->Release(pFileDialog);
ecreate:
#else // ] [
  // https://www.codeproject.com/Articles/13088/How-to-Browse-for-a-Folder
  BROWSEINFOW bi={ 0 };
  LPITEMIDLIST pIdl;

  bi.ulFlags=BIF_USENEWUI;
  bi.lpszTitle = aTitle;
  pIdl=SHBrowseForFolderW(&bi);

  if (!pIdl)
    goto exit;

  SHGetPathFromIDListW(pIdl,path);
  CoTaskMemFree(pIdl);
  code=0;
exit:
#endif // ]
  return code;
}

void GtkPathSave(Gtk *pGtk)
{
  if (pGtk->pPlayer->ini[0]) {
    PropertiesWrite(gcaPropertiesGtk,pGtk->pPlayer->ini,gcaKeyGtk,
        &pGtk->options.current,&pGtk->options.original);
    pGtk->options.original=pGtk->options.current;
  }
}

static DWORD WINAPI GtkThreadProc(LPVOID lpParameter)
{
  Gtk *pGtk=lpParameter;
  Dialog *pDialogMax=pGtk->aDialog+eGtkNumDialogs;
  Dialog *pDialog;
  GtkThread thread;
  int code;

  /////////////////////////////////////////////////////////////////////////////
  ++pGtk->eState;
  DWRITELN(GtkState(pGtk));
  DASSERT(eGtkThreadRunning==pGtk->eState,exit);
  SetEvent(pGtk->hProducer);

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&thread,sizeof thread);
  thread.pGtk=pGtk;

  if (GtkThreadSubclass(&thread)<0) {
    DMESSAGE("subclassing");
    goto esubclass;
  }

  /////////////////////////////////////////////////////////////////////////////
  for (;;) {
    code=GtkThreadCreate(&thread);

    if (0<code) {
      if (GtkThreadShowPathDialog(&thread)<0) {
        DMESSAGE("selecting path");
        GtkThreadUnsubclass(&thread);
        goto ethread;
      }

      GtkPathSave(pGtk);
    }
    else if (code<0) {
      DMESSAGE("creating thread");
      GtkThreadUnsubclass(&thread);
      goto ethread;
    }
    else
      break;
  }

  for (pDialog=pGtk->aDialog;pDialog<pDialogMax;++pDialog) {
    pDialog->lpVtbl->WindowCreate (pDialog);
    DASSERT (GTK_IS_WIDGET (pDialog->window), exit);
    gtk_window_set_title (GTK_WINDOW (pDialog->window),
        pDialog->lpVtbl->pTitle);
    gtk_window_set_default_size (GTK_WINDOW (pDialog->window),
        pDialog->lpVtbl->nWidthPixels, -1);
    gtk_window_set_position (GTK_WINDOW(pDialog->window),
        GTK_WIN_POS_CENTER);
    g_signal_connect (G_OBJECT(pDialog->window), "delete-event",
        G_CALLBACK(window_delete_event), pDialog);
  }

  if (pGtk->pDialog)
    GtkDialogShowSync (pGtk->pDialog,0);

  GtkThreadUnsubclass(&thread);

  if (pGtk->pDialog)
    GtkDialogShowSync (pGtk->pDialog,1);

  ++pGtk->eState;
  DWRITELN(GtkState(pGtk));
  DASSERT(eGtkLoaded==pGtk->eState,exit);
  gtk_main();
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  GtkThreadDestroy(&thread);
ethread:
esubclass:
  DWRITELN("END WIN THREAD!!!");
  ExitThread(0ul);
  return 0ul;
}

static gboolean ConfigWidgetUpdateAsync(gpointer user_data)
{
  ConfigWidget *pWidget=user_data;

  pWidget->lpVtbl->set(pWidget);

  return FALSE;
}

void GtkUpdateWidget(Gtk *pGtk, ConfigWidget *pWidget)
{
  if (pWidget->lpVtbl&&eGtkLoaded<=pGtk->eState)
    g_idle_add (ConfigWidgetUpdateAsync, pWidget);
}

static gboolean DialogUpdateAsync(gpointer user_data)
{
  Dialog *pDialog=user_data;

  pDialog->lpVtbl->set(pDialog);

  return FALSE;
}

void GtkUpdateDialog(Gtk *pGtk, Dialog *pDialog)
{
  if (pDialog->lpVtbl&&eGtkLoaded<=pGtk->eState)
    g_idle_add (DialogUpdateAsync, pDialog);
}

#if defined (YASAPI_DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
static const char *GtkThreadState(const GtkThread *pThread)
{
  const char *mnemonic[1+eGtkThreadMax]={
    "undefined",
    "eGtkThreadNull",
#if defined (YASAPI_GTK_LOAD) // [
    "eGtkThreadGtkLoaded",
#endif // ]
    "eGtkThreadDialogs",
  };

  EGtkThread eState=pThread->eState;

  return mnemonic[0<=eState&&eState<eGtkThreadMax?1+eState:0];
}

const char *GtkState(const Gtk *pGtk)
{
  const char *mnemonic[1+eGtkMax]={
    "undefined",
    "eGtkNull",
    "eGtkOptions",
    "eGtkCriticalSection",
    "eGtkProducer",
    "eGtkConsumer",
    "eGtkThread",
    "eGtkThreadRunning",
    "eGtkLoaded",
  };

  EGtk eState=pGtk->eState;

  return mnemonic[0<=eState&&eState<eGtkMax?1+eState:0];
}
#endif // ]
