/*
 * yasapi_options.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
const wchar_t gcaKeyGtk[]=L"Gtk";

const OptionsGtk gcOptionsGtk={
  .path=L"",
};

///////////////////////////////////////////////////////////////////////////////
const OptionsDialog gcOptionsDialog={
  .nPosX=-1,
  .nPosY=-1,
  .nPage=-1,
};

///////////////////////////////////////////////////////////////////////////////
static const wchar_t gcaKeyCommon[]=L"Common";

static void OptionsMemberDeviceRollback(OptionsMemberDevice *pDevice,
    Options *pOptions);
static void OptionsMemberDeviceSave(OptionsMemberDevice *pDevice,
    Options *pOptions);

const OptionsCommon gcOptionsCommon={
  .id=L"",
  .bLatency=0,
  .eMono2Stereo=eMono2StereoPlain,
  .eVolume=eVolumeSoftware,
  .bSurround=0,
  .bDisconnect=1,
#if defined (YASAPI_OPTION_GAPLESS) // [
  .bGapless=1,
#else // ] [
  .bGapless=0,
#endif // ]
  .pos.eConfig=eDialogPosOnTopPreferences,
  .pos.eAbout=eDialogPosOnTopPreferences,
};

///////////////////////////////////////////////////////////////////////////////
const OptionsDevice gcOptionsDevice={
  .eMode=eModeShared,
  .bAutoConvertPCM=1,
  .bSRCDefaultQuality=0,
  .eStrategy=eStrategyPush,
  .bTimerPeriodic=1,
  .bAlignmentDance=1,
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  .bProAudio=1,
#endif // ]
  .ePeriod=ePeriodDefault,
  .eBitdepth=eBitdepthDisabled,
  .xRingSize=16.0,
  .xMaxLatencyAssumed=500.0,
  .xPlaybackBegin=2.0,
};

///////////////////////////////////////////////////////////////////////////////
int OptionsCreate(Options *pOptions, Player *pPlayer, UINT cDevices)
{
  /////////////////////////////////////////////////////////////////////////////
  pOptions->ini=pPlayer->ini;

  /////////////////////////////////////////////////////////////////////////////
  if (PlayerPropertiesPath(pPlayer)<0)
    goto properties;

  /////////////////////////////////////////////////////////////////////////////
properties:
  PropertiesRead(gcaPropertiesCommon,pOptions->ini,gcaKeyCommon,
      &pOptions->common.original,&gcOptionsCommon);
  pOptions->common.current=pOptions->common.original;
//DVWRITELNW(L"id=\"%s\"",pOptions->common.current.id);
  pOptions->common.config=pOptions->common.current;

  /////////////////////////////////////////////////////////////////////////////
  pOptions->pDevices=malloc(cDevices*sizeof *pOptions->pDevices);

  if (!pOptions->pDevices) {
    DMESSAGE("allocating devices");
    goto edevices;
  }

  pOptions->pDevicesMax=pOptions->pDevices;
  pOptions->pDevicesAlloc=pOptions->pDevices+cDevices;

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//error:
  free(pOptions->pDevices);
edevices:
  return -1;
}

void OptionsDestroy(Options *pOptions)
{
  OptionsMemberDevice *pDevice=pOptions->pDevicesMax;

  while (pOptions->pDevices<pDevice)
    OptionsMemberDeviceUnref(--pDevice,pOptions);

  free(pOptions->pDevices);

DVWRITELNW(L"*** destroy: id=\"%s\"",pOptions->common.current.id);
  PropertiesWrite(gcaPropertiesCommon,pOptions->ini,gcaKeyCommon,
      &pOptions->common.current,&pOptions->common.original);
}

OptionsMemberDevice *OptionsSearch(Options *pOptions, const wchar_t *pId)
{
  OptionsMemberDevice *pDevice=pOptions->pDevices;

  while (pDevice<pOptions->pDevicesMax) {
    if (!wcscmp(pId,pDevice->pId))
      goto success;

    ++pDevice;
  }

  if (pOptions->pDevicesMax==pOptions->pDevicesAlloc) {
    DMESSAGE("overflow");
    goto eoverflow;
  }

  ++pOptions->pDevicesMax;
  PropertiesRead(gcaPropertiesDevice,pOptions->ini,pId,
      &pDevice->original,&gcOptionsDevice);
  pDevice->current=pDevice->original;
  pDevice->config=pDevice->current;
  pDevice->pId=pId;
  pDevice->uRef=0u;
success:
  return pDevice;
//error:
eoverflow:
  return NULL;
}

void OptionsRollback(Options *pOptions)
{
  OptionsMemberDevice *pDevice=pOptions->pDevicesMax;

  while (pOptions->pDevices<pDevice)
    OptionsMemberDeviceRollback(--pDevice,pOptions);

  /*
  PropertiesWrite(gcaPropertiesCommon,pOptions->ini,gcaKeyCommon,
      &pOptions->common.current,&pOptions->common.original);
  */
  pOptions->common.config=pOptions->common.current;
}

void OptionsSave(Options *pOptions)
{
  OptionsMemberDevice *pDevice=pOptions->pDevicesMax;

  while (pOptions->pDevices<pDevice)
    OptionsMemberDeviceSave(--pDevice,pOptions);

  PropertiesWrite(gcaPropertiesCommon,pOptions->ini,gcaKeyCommon,
      &pOptions->common.current,&pOptions->common.original);
  pOptions->common.original=pOptions->common.current;
}

///////////////////////////////////////////////////////////////////////////////
void OptionsMemberDeviceRef(OptionsMemberDevice *pDevice)
{
  ++pDevice->uRef;
}

void OptionsMemberDeviceUnref(OptionsMemberDevice *pDevice,
    Options *pOptions)
{
  if (pDevice->uRef&&!--pDevice->uRef) {
    PropertiesWrite(gcaPropertiesDevice,pOptions->ini,pDevice->pId,
        &pDevice->current,&pDevice->original);
  }
}

static void OptionsMemberDeviceRollback(OptionsMemberDevice *pDevice,
    Options *pOptions)
{
  /*
  PropertiesWrite(gcaPropertiesDevice,pOptions->ini,pDevice->pId,
      &pDevice->current,&pDevice->original);
  */
  pDevice->config=pDevice->current;
}

static void OptionsMemberDeviceSave(OptionsMemberDevice *pDevice,
    Options *pOptions)
{
  PropertiesWrite(gcaPropertiesDevice,pOptions->ini,pDevice->pId,
      &pDevice->current,&pDevice->original);
  pDevice->original=pDevice->current;
}
