/*
 * yasapi_param.h
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef __YASAPI_PARAM_H__ // [
#define __YASAPI_PARAM_H__

#include <yasapi_defines.h>
#include <windows.h>
#include <mmdeviceapi.h>
#include <audiosessiontypes.h>
#include <audioclient.h>
#include <functiondiscoverykeys_devpkey.h>
#if defined (YASAPI_AUDIO_ENDPOINT_VOLUME) // [
#include <endpointvolume.h>
#endif // ]
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
#include <shobjidl.h>
#else // ] [
#include <shlobj.h>
#endif // ]
#include <gtk/gtk.h>
#include <gdk/gdkwin32.h>

#ifdef __cplusplus // [
extern "C" {
#endif // ]

///////////////////////////////////////////////////////////////////////////////
extern const CLSID CLSID_MMDeviceEnumerator;
extern const IID IID_IMMDeviceEnumerator;
extern const IID IID_IAudioClient;
extern const IID IID_IAudioClock;
extern const IID IID_IAudioRenderClient;
#if defined (YASAPI_SIMPLE_AUDIO_VOLUME) // [
extern const IID IID_ISimpleAudioVolume;
#endif // ]
#if defined (YASAPI_AUDIO_ENDPOINT_VOLUME) // [
extern const IID IID_IAudioEndpointVolume;
#endif // ]
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
extern const CLSID CLSID_FileOpenDialog;
extern const IID IID_IFileDialog;
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef struct _Player Player;
typedef struct _Dialog Dialog;

typedef struct _OptionsGtk OptionsGtk;
typedef struct _OptionsDialog OptionsDialog;
typedef enum _EMono2Stereo EMono2Stereo;
typedef enum _EDialogPos EDialogPos;
typedef enum _EVolume EVolume;
typedef struct _OptionsCommon OptionsCommon;
typedef enum _EPeriod EPeriod;
#define YASAPI_MODE
#if defined (YASAPI_MODE) // [
typedef enum _EMode EMode;
#endif // ]
#define YASAPI_STRATEGY
#if defined (YASAPI_STRATEGY) // [
typedef enum _EStrategy EStrategy;
#endif // ]
typedef enum _EBitdepth EBitdepth;
typedef struct _OptionsDevice OptionsDevice;

typedef struct _OptionsMemberGtk OptionsMemberGtk;
typedef struct _OptionsMemberCommon OptionsMemberCommon;
typedef struct _OptionsMemberDevice OptionsMemberDevice;
typedef struct _Options Options;

typedef const struct _IWasapiDeviceVtbl IWasapiDeviceVtbl;
typedef struct _WasapiDevice WasapiDevice;
typedef struct _Wasapi Wasapi;

typedef const struct _IPropertyVtbl IPropertyVtbl;
typedef struct _Property Property;

typedef enum _EFormatMatrixLimits EFormatMatrixLimits;
typedef const struct _Frequency Frequency;
typedef const struct _BitDepth BitDepth;
typedef struct _FormatMatrix FormatMatrix;
typedef const struct _IConfigWidgetVtbl IConfigWidgetVtbl;
typedef struct _ConfigWidget ConfigWidget;
typedef char *Navigate(Dialog *pDialog, int offs);
typedef struct _ConfigGeneral ConfigGeneral;
typedef struct _ConfigCommon ConfigCommon;
typedef struct _ConfigWasapi ConfigWasapi;
typedef struct _ConfigYasapi ConfigYasapi;
typedef struct _ConfigMonitor ConfigMonitor;
typedef struct _ConfigFormatMatrix ConfigFormatMatrix;
typedef const struct _ClassCell ClassCell;

typedef const struct _IDialogVtbl IDialogVtbl;
typedef struct _Config Config;
typedef enum _EAboutLimits EAboutLimits;
typedef struct _About About;
typedef enum _EDialog EDialog;

typedef enum _EGtk EGtk;
typedef enum _EGtkLimits EGtkLimits;
typedef struct _Gtk Gtk;
typedef struct _GtkDialogs GtkDialogs;

///////////////////////////////////////////////////////////////////////////////
enum _EFormatMatrixLimits {
  eFrequenciesNum=8,
  eBitDepthsNum=3,
};

struct _FormatMatrix {
  int bSupported[eFrequenciesNum][eBitDepthsNum];
};

int FormatMatrixCreate(FormatMatrix *pMatrix, IMMDevice *pDevice);
void FormatMatrixDestroy(FormatMatrix *pMatrix);

///////////////////////////////////////////////////////////////////////////////
struct _OptionsGtk {
  wchar_t path[MAX_PATH];
};

extern const wchar_t gcaKeyGtk[];
extern const OptionsGtk gcOptionsGtk;

////////////////
struct _OptionsDialog {
  int nPosX;
  int nPosY;
  int nPage;
};

extern const OptionsDialog gcOptionsDialog;

////////////////
enum _EMono2Stereo {
  eMono2StereoOff,
  eMono2StereoPlain,
  eMono2StereoHalf,
  eMono2StereoSqrt2Inverse,
};

enum _EDialogPos {
  eDialogPosFreeFloating,
  eDialogPosOnTopPreferences,
  eDialogPosAlwaysOnTop,
};

////////////////
enum _EVolume {
  eVolumeOff,
  eVolumeSoftware,
#if defined (YASAPI_VOLUME_WASAPI) // [
  eVolumeWasapi,
#endif // ]
};

struct _OptionsCommon {
  wchar_t id[128];
  int bLatency;
  EMono2Stereo eMono2Stereo;
  EVolume eVolume;
  int bSurround;
  int bDisconnect;
  int bGapless;

  struct {
    EDialogPos eConfig;
    EDialogPos eAbout;
  } pos;
};

extern const OptionsCommon gcOptionsCommon;

////////////////
enum _EBitdepth {
  eBitdepthDisabled=0,
  eBitdepth16=16,
  eBitdepth24=24,
  eBitdepth32=32,
};

enum _EPeriod {
  ePeriodDefault,
  ePeriodMinimum,
};

enum _EMode {
  eModeShared,
  eModeExclusive,
};

enum _EStrategy {
  eStrategyPush,
  eStrategyPull,
};

struct _OptionsDevice {
  EMode eMode;
  int bAutoConvertPCM;
  int bSRCDefaultQuality;
  EStrategy eStrategy;
  int bTimerPeriodic;
  int bAlignmentDance;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  int bProAudio;
#endif // ]
  EBitdepth eBitdepth;
  EPeriod ePeriod;
  double xRingSize;
  double xMaxLatencyAssumed;
  double xPlaybackBegin;
};

extern const OptionsDevice gcOptionsDevice;

////////////////
struct _OptionsMemberGtk {
  OptionsGtk config;
  OptionsGtk current;
  OptionsGtk original;
};

////////////////
struct _OptionsMemberCommon {
  OptionsCommon config;
  OptionsCommon current;
  OptionsCommon original;
};

////////////////
struct _OptionsMemberDevice {
  const wchar_t *pId;
  UINT uRef;
  OptionsDevice config;
  OptionsDevice current;
  OptionsDevice original;
  FormatMatrix matrix;
};

OptionsMemberDevice *OptionsSearch(Options *pOptions, const wchar_t *id);
void OptionsMemberDeviceRef(OptionsMemberDevice *pDevice);
void OptionsMemberDeviceUnref(OptionsMemberDevice *pDevice,
    Options *pOptions);

////////////////
struct _Options {
  wchar_t *ini;
  OptionsMemberCommon common;
  OptionsMemberDevice *pDevices;
  OptionsMemberDevice *pDevicesMax;
  OptionsMemberDevice *pDevicesAlloc;
};

int OptionsCreate(Options *pOptions, Player *pPlayer, UINT cDevices);
void OptionsDestroy(Options *pOptions);

void OptionsRollback(Options *pOptions);
void OptionsSave(Options *pOptions);

///////////////////////////////////////////////////////////////////////////////
struct _IWasapiDeviceVtbl {
#if defined (YASAPI_DEBUG) // [
  const char *pId;
#endif // ]
  const wchar_t *(*getid)(WasapiDevice *pDevice);
};

struct _WasapiDevice {
  IWasapiDeviceVtbl *lpVtbl;
  IMMDevice *pDevice;
  wchar_t *id;
  IPropertyStore *pStore;
  PROPVARIANT vFriendlyName;
  char *pFriendlyNameOriginal;
  char *pFriendlyName;
#if defined (YASAPI_DEVICE_MARKUP) // [
  char *pFriendlyNameMarkup;
#endif // ]
  OptionsMemberDevice *pMember;
};

int WasapiDeviceCreate(WasapiDevice *pDevice, Wasapi *pWasapi, int nDevice);
void WasapiDeviceDestroy(WasapiDevice *pDevice, Options *pOptions);

///////////////////////////////////////////////////////////////////////////////
struct _Wasapi {
  IMMDeviceEnumerator *pEnumerator;
  IMMDeviceCollection *pCollection;
  Options options;

  struct {
    WasapiDevice *pDevice;
    WasapiDevice *pDeviceConfig;
  } selected;

  UINT cDevices;
  WasapiDevice *pDevices;
};

int WasapiCreate(Wasapi *pWasapi, Player *pPlayer);
void WasapiDestroy(Wasapi *pWasapi, Player *pPlayer);

int WasapiSelectDevice(Wasapi *pWasapi, WasapiDevice *pDevice);
void WasapiRestore(Wasapi *pWasapi);
void WasapiSync(Wasapi *pWasapi);
void WasapiDefault(Wasapi *pWasapi);

///////////////////////////////////////////////////////////////////////////////
struct _IPropertyVtbl {
  void (*read)(const Property *pProperty, const wchar_t *path,
      const wchar_t *lpAppName, char *data, const char *ref);
  void (*write)(const Property *pProperty, const wchar_t *path,
      const wchar_t *lpAppName, const char *data, const char *ref);
};

struct _Property {
  IPropertyVtbl *lpVtbl;
  const wchar_t *lpKeyName;
  int offs;
  DWORD dwSize;
};

extern const Property gcaPropertiesDialog[];
extern const Property gcaPropertiesGtk[];
extern const Property gcaPropertiesCommon[];
extern const Property gcaPropertiesDevice[];

void PropertiesRead(const Property *pProperty, const wchar_t *path,
    const wchar_t *lpAppName, void *data, const void *ref);
void PropertiesWrite(const Property *pProperty, const wchar_t *path,
    const wchar_t *lpAppName, const void *data, const void *ref);

///////////////////////////////////////////////////////////////////////////////
struct _Frequency {
  const char *label;
  int freq;
};

extern Frequency gaFrequencies[eFrequenciesNum];

struct _BitDepth {
  const char *label;
  int bits;
};

extern BitDepth gaBitDepths[eBitDepthsNum];

struct _IConfigWidgetVtbl {
#if defined (YASAPI_DEBUG) // [
  const char *pId;
#endif // ]
  void (*set)(ConfigWidget *pWidget);
  void (*get)(ConfigWidget *pWidget);
};

struct _ConfigWidget {
  IConfigWidgetVtbl *lpVtbl;
  Dialog *pDialog;
  ClassCell *pCell;
  GtkWidget *widget;
};

struct _ConfigGeneral {
  ConfigWidget device;
  ConfigWidget mode;
  ConfigWidget strategy;
};

struct _ConfigCommon {
  ConfigWidget latency;
  ConfigWidget mono2stereo;
  ConfigWidget volume;
  ConfigWidget surround;
  ConfigWidget disconnect;
  ConfigWidget gapless;

  struct {
    ConfigWidget config;
    ConfigWidget about;
  } pos;
};

struct _ConfigWasapi {
  ConfigWidget device;
  ConfigWidget mode;
  ConfigWidget auto_convert_pcm;
  ConfigWidget src_default_quality;
  ConfigWidget strategy;
  ConfigWidget timer_periodic;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  ConfigWidget pro_audio;
#endif // ]
  ConfigWidget alignment_dance;
  ConfigWidget period;
};

struct _ConfigYasapi {
  ConfigWidget device;
  ConfigWidget bitdepth;
  ConfigWidget ring_size;
  ConfigWidget playback_begin;
  ConfigWidget period_playback_begin;
  ConfigWidget max_latency_assumed;
};

struct _ConfigMonitor {
  ConfigWidget device;

  struct {
    ConfigWidget num_tracks;
    ConfigWidget num_gapless;
    ConfigWidget num_dropouts;
    ConfigWidget ms_latency;
  } player;

  struct {
#if defined (YASAPI_OPTION_VOLUME) // [
    ConfigWidget volume;
#endif // ]
    ConfigWidget num_tracks;
    ConfigWidget num_frames_device;
    ConfigWidget num_frames_ring;
    ConfigWidget num_frames_latency;
    ConfigWidget num_dropouts;
  } engine;

  ConfigWidget written;
  ConfigWidget padding;
};

struct _ConfigFormatMatrix {
  ConfigWidget device;
  ConfigWidget enabled[eFrequenciesNum][eBitDepthsNum];
};

///////////////////////////////////////////////////////////////////////////////
struct _IDialogVtbl {
  const wchar_t *pIdW;
#if defined (YASAPI_DEBUG) // [
  const char *pIdA;
#endif // ]
  const char *pTitle;
  int nWidthPixels;
  void (*WindowCreate)(Dialog *pDialog);
  void (*set)(Dialog *pDialog);
  void (*get)(Dialog *pDialog);
};

struct _Config {
  union {
    struct {
      ConfigGeneral general;
      ConfigCommon common;
      ConfigWasapi wasapi;
      ConfigYasapi yasapi;
      ConfigMonitor monitor;
      ConfigFormatMatrix explore;
    };

    ConfigWidget widgets[1];
  };
};

enum _EAboutLimits {
  eAboutLimitsURLSize = 128,
  eAboutLimitsLinksSize = 30,
  eAboutLimitsPagesNum = 3,
  //eAboutLimitsWidthPixels=455,
  eAboutLimitsWidthPixels=492,
  //eAboutLimitsWidthPixels=518,
  eAboutLimitsHeightPixels=296,
  eAboutLimitsWidthChars=77,
};

struct _About {
#if defined (YASAPI_ABOUT_TEXT_VIEW) // [
  GdkDisplay *display;
  GdkCursor *arrow;
  GdkCursor *hand;
  AboutPage *pPageActive;
  AboutPage pages[eAboutLimitsPagesNum];
#else // ] [
  int dummy;
#endif // ]
};

extern IDialogVtbl gIAboutVtbl;
extern IDialogVtbl gIConfigVtbl;

enum _EDialog {
  eAboutWidthPixels=492,
  eConfigWidthPixels=680,
};

struct _Dialog {
  IDialogVtbl *lpVtbl;
  HWND hWndParent;
  Gtk *pGtk;
  GtkWidget *window;
  GtkWidget *notebook;
  OptionsDialog options;
  int bMoved;

  union {
    About about;
    Config config;
  };
};

////////////////
enum _EGtk {
  eGtkNull,
  eGtkOptions,
  eGtkCriticalSection,
  eGtkProducer,
  eGtkConsumer,
  eGtkThread,
  eGtkThreadRunning,
  eGtkLoaded,
  //////////////
  eGtkMax,
};

void gtk_dialog_options_save (GtkWidget *widget, gpointer data);

////////////////
enum _EGtkLimits {
  eGtkNumDialogs=2,
};

struct _Gtk {
  EGtk eState;
  Player *pPlayer;
  OptionsMemberGtk options;
  Dialog *pDialog;

  union {
    struct {
      Dialog about;
      Dialog config;
    };

    Dialog aDialog[eGtkNumDialogs];
  };

  CRITICAL_SECTION cs;
  HMODULE hLibGlib;
  HANDLE hProducer;
  HANDLE hConsumer;
  HANDLE hThread;
  GtkCssProvider *provider;
};

int GtkShow(Gtk *pGtk, Player *pPlayer, Dialog *pDialog, HWND hWndParent);
void GtkDestroy(Gtk *pGtk);

void GtkPathSave(Gtk *pGtk);
void GtkUpdateWidget(Gtk *pGtk, ConfigWidget *pWidget);
void GtkUpdateDialog(Gtk *pGtk, Dialog *pDialog);

#if defined (YASAPI_DEBUG) // [
const char *GtkState(const Gtk *pGtk);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
char *yasapi_wcs2strdup(wchar_t *wcs, UINT uCodePage);
void yasapi_message(const char *file, int line, const char *format,
    DWORD dwError, ...);
const char *yasapi_basenamea(const char *path);

#ifdef __cplusplus // [
}
#endif // ]
#endif // __YASAPI_PARAM_H__ ]
