/*
 * yasapi_player.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <winamp/wa_ipc.h>

///////////////////////////////////////////////////////////////////////////////
// http://forums.winamp.com/archive/index.php/t-316807.html
#define IPC_GETINIFILEW 1334
#define IPC_GETINIDIRECTORYW 1335
#define IPC_GETPLUGINDIRECTORYW 1336
#define IPC_GETVISDIRECTORYW 339
#define IPC_GETSKINDIRECTORYW 340
#define IPC_GETDSPDIRECTORYW 341

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, Out_Module *pOut)
{
  int code=-1;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(pPlayer,sizeof *pPlayer);
  pPlayer->pOut=pOut;
  pPlayer->ini[0]=0;
#if defined (YASAPI_OPTION_VOLUME) // [
  pPlayer->params.nVolume=ePlayerVolumeMax;
#endif // ]
  pPlayer->stats.uMsLatencyMax=0u;

#if defined (YASAPI_DEBUG) // [
  /////////////////////////////////////////////////////////////////////////////
  if (TraceCreate(&pPlayer->trace,pOut)<0)
    goto exit;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return code;
}

void PlayerDestroy(Player *pPlayer)
{
  GtkDestroy(&pPlayer->gtk);
  EngineDestroy(&pPlayer->engine);
#if defined (YASAPI_DEBUG) // [
  TraceDestroy(&pPlayer->trace);
#endif // ]
  if (pPlayer->wasapi.pEnumerator)
  	WasapiDestroy(&pPlayer->wasapi,pPlayer);
}

///////////////////////////////////////////////////////////////////////////////
int PlayerPropertiesPath(Player *pPlayer)
{
  static const wchar_t SUBDIR[]=L"\\Plugins";

  int code=-1;
  Out_Module *pOut=pPlayer->pOut;
  wchar_t *ini=pPlayer->ini;
  wchar_t  path[MAX_PATH];
  DWORD dwSize;
  wchar_t *inidir,*ep,*pp,*wp;

  /////////////////////////////////////////////////////////////////////////////
  ini[0]=0;

  if (!pOut)
    goto eout;

  /////////////////////////////////////////////////////////////////////////////
  dwSize=GetModuleFileNameW(
    pOut->hDllInstance, // _In_opt_ HMODULE hModule,
    path,               // _Out_    LPTSTR  lpFilename,
    sizeof path         // _In_     DWORD   nSize
  );

  // extension pointer:
  for (ep=path+dwSize;path<ep&&L'.'!=ep[-1];--ep)
    ;

  // path pointer:
  for (pp=ep;path<pp&&L'\\'!=pp[-1];--pp)
    ;

  // write pointer:
  wp=ini;

  // copy the path.
  inidir=(wchar_t *)SendMessageW(pOut->hMainWindow,WM_WA_IPC,0,
      IPC_GETINIDIRECTORYW);
  wcscpy(wp,inidir);
  wp+=wcslen(inidir);
  wcscpy(wp,SUBDIR);
  wp+=wcslen(SUBDIR);

  CreateDirectoryW(
    ini,      // _In_     LPCTSTR               lpPathName,
    NULL      // _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes
  );

  *wp++=L'\\';
  *wp=0;

  // copy the basename without extension.
  memcpy(wp,pp,(ep-pp)*sizeof *wp);
  wp+=ep-pp;
  *wp=0;

  // copy the extension.
  wcscpy(wp,L"ini");

  /////////////////////////////////////////////////////////////////////////////
  code=0;
eout:
  return code;
}

int PlayerOpen(Player *pPlayer, int samplerate, int numchannels,
    int bitspersamp)
{
  int code=-1;

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreate(&pPlayer->engine,pPlayer,samplerate,numchannels,
      bitspersamp)<0) {
    DMESSAGE("creating engine");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

void PlayerClose(Player *pPlayer)
{
DMARKLN();
  EngineClose(&pPlayer->engine);
DMARKLN();
}

Wasapi *PlayerWasapi(Player *pPlayer)
{
  if (pPlayer->wasapi.pEnumerator)
    goto success;

  if (WasapiCreate(&pPlayer->wasapi,pPlayer)<0) {
    DMESSAGE("Creating wasapi");
    goto ewasapi;
  }
success:
  return &pPlayer->wasapi;
//error:
  WasapiDestroy(&pPlayer->wasapi,pPlayer);
ewasapi:
  return NULL;
}

void PlayerFlush(Player *pPlayer)
{
  Engine *pEngine=&pPlayer->engine;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  if (pEngine->eState<eEngineStarted)
    goto exit;

  EngineSend(pEngine,eCommandFlush,eEngineNull,0);
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return;
}

int PlayerPause(Player *pPlayer, int nPause)
{
  return EnginePause(&pPlayer->engine,nPause);
}

int PlayerIsPlaying(Player *pPlayer)
{
#if 0 && defined (YASAPI_DEBUG) // [
  int bIsPlaying;

  bIsPlaying=EngineIsPlaying(&pPlayer->engine);

  return bIsPlaying;
#else // ] [
  return EngineIsPlaying(&pPlayer->engine);
#endif // ]
}

int PlayerAvailable(Player *pPlayer)
{
  return EngineAvailable(&pPlayer->engine);
}

int PlayerOutputTime(Player *pPlayer)
{
  return EngineOutputTimeExt(&pPlayer->engine);
}

int PlayerWrittenTime(Player *pPlayer)
{
  return EngineWrittenTime(&pPlayer->engine);
}

int PlayerWrite(Player *pPlayer, BYTE *buf, int len)
{
  return EngineWrite(&pPlayer->engine,buf,len);
}

int PlayerShow(Player *pPlayer, Dialog *pDialog, HWND hWndParent)
{
  return GtkShow(&pPlayer->gtk,pPlayer,pDialog,hWndParent);
}

#if defined (YASAPI_OPTION_VOLUME) // [
void PlayerVolume(Player *pPlayer, int nVolume)
{
  if (0<=nVolume) {
    pPlayer->params.nVolume=nVolume;
    EngineVolume(&pPlayer->engine);
  }
}
#endif // ]
