/*
 * yasapi_ring_m2s.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#if 0 && defined (YASAPI_FORMAT_CONVERT) // [
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
static DWORD RingMono2StereoConvertSizeSource(Ring *pRing, DWORD dwSize)
{
  return dwSize>>1ul;
}

static DWORD RingMono2StereoSizeSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_mp-pRing->_buf);
}

static DWORD RingMono2StereoWrittenSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_dwWritten);
}

static DWORD RingMono2StereoAvailableSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_dwAvailable);
}

////////////////
static DWORD RingMono2StereoConvertSizeTarget(Ring *pRing, DWORD dwSize)
{
  return dwSize<<1ul;
}

static DWORD RingMono2StereoSizeTarget(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

static DWORD RingMono2StereoWrittenTarget(Ring *pRing)
{
  return pRing->_dwWritten;
}

#if defined (YASAPI_TARGET_AVAILABLE) // [
static DWORD RingMono2StereoAvailableTarget(Ring *pRing)
{
  return pRing->_dwAvailable;
}
#endif // ]

////////////////
static double xMono2StereoSqrt2Inverse=0.0;

static BYTE *RingMono2StereoCopyToRing(Ring *pRing, BYTE *pbDst,
    const BYTE *pbSrc, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WORD nBlockAlignSource=pEngine->wfx.source.Format.nBlockAlign;
  BYTE *pbDstMax=pbDst+dwSize;
  INT32 i;
  BYTE *ip=(BYTE *)&i+(sizeof i)-nBlockAlignSource;

  switch (pEngine->options.common.eMono2Stereo) {
  case eMono2StereoHalf:
    while (pbDst<pbDstMax) {
      i=0;
      memcpy(ip,pbSrc,nBlockAlignSource);
      //i=0.5*i+0.5;
      i>>=1;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  case eMono2StereoSqrt2Inverse:
    if (0.0==xMono2StereoSqrt2Inverse)
      xMono2StereoSqrt2Inverse=1.0/sqrt(2.0);

    while (pbDst<pbDstMax) {
      i=0;
      memcpy(ip,pbSrc,nBlockAlignSource);
      i=xMono2StereoSqrt2Inverse*i+0.5;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  default:
    while (pbDst<pbDstMax) {
      memcpy(pbDst,pbSrc,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,pbSrc,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  }

  return pbDst;
}

static BYTE *RingMono2StereoCopyFromRing(Ring *pRing,
    BYTE *pbDst, const BYTE *pbSrc, DWORD dwSize)
{
  return memcpy(pbDst,pbSrc,dwSize);
}

////////////////
IRingVtbl gcIRingMono2StereoVtbl={
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
  .source.ConvertSize=RingMono2StereoConvertSizeSource,
#endif // ]
  .source.Size=RingMono2StereoSizeSource,
  .source.Written=RingMono2StereoWrittenSource,
  .source.Available=RingMono2StereoAvailableSource,
  ////////////
  .target.ConvertSize=RingMono2StereoConvertSizeTarget,
  .target.Size=RingMono2StereoSizeTarget,
  .target.Written=RingMono2StereoWrittenTarget,
#if defined (YASAPI_TARGET_AVAILABLE) // [
  .target.Available=RingMono2StereoAvailableTarget,
#endif // ]
  ////////////
  .CopyToRing=RingMono2StereoCopyToRing,
  .CopyFromRing=RingMono2StereoCopyFromRing,
};
#endif // ]
