/*
 * yasapi_ring_conv.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <stdint.h>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Transform

#define YASAPI_ID_IRING_VTBL() \
  YASAPI_PASTE3(gcIRing,YASAPI_ID,Vtbl)
#define YASAPI_ID_RING(id) \
  YASAPI_PASTE3(Ring,YASAPI_ID,id)
#define YASAPI_ID_RING_SOURCE(id) \
  YASAPI_PASTE4(Ring,YASAPI_ID,Source,id)
#define YASAPI_ID_RING_TARGET(id) \
  YASAPI_PASTE4(Ring,YASAPI_ID,Target,id)

///////////////////////////////////////////////////////////////////////////////
static DWORD YASAPI_ID_RING_SOURCE(ConvertSize)(Ring *pRing, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;

  dwSize=MulDiv(dwSize,pSource->wBitsPerSample,pTarget->wBitsPerSample);

  if (!pEngine->options.common.eMono2Stereo)
    goto exit;

  if (1u!=pSource->nChannels||2u!=pTarget->nChannels)
    goto exit;

  dwSize>>=1ul;
exit:
  return dwSize;
}

static DWORD YASAPI_ID_RING_SOURCE(Size)(Ring *pRing)
{
  return YASAPI_ID_RING_SOURCE(ConvertSize)(pRing,pRing->_mp-pRing->_buf);
}

static DWORD YASAPI_ID_RING_SOURCE(Written)(Ring *pRing)
{
  return YASAPI_ID_RING_SOURCE(ConvertSize)(pRing,pRing->_dwWritten);
}

static DWORD YASAPI_ID_RING_SOURCE(Available)(Ring *pRing)
{
  return YASAPI_ID_RING_SOURCE(ConvertSize)(pRing,pRing->_dwAvailable);
}

////////////////
static DWORD YASAPI_ID_RING_TARGET(ConvertSize)(Ring *pRing, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;

  dwSize=MulDiv(dwSize,pTarget->wBitsPerSample,pSource->wBitsPerSample);

  if (!pEngine->options.common.eMono2Stereo)
    goto exit;

  if (1u!=pSource->nChannels||2u!=pTarget->nChannels)
    goto exit;

  dwSize<<=1ul;
exit:
  return dwSize;
}

static DWORD YASAPI_ID_RING_TARGET(Size)(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

static DWORD YASAPI_ID_RING_TARGET(Written)(Ring *pRing)
{
  return pRing->_dwWritten;
}

#if defined (YASAPI_TARGET_AVAILABLE) // [
static DWORD YASAPI_ID_RING_TARGET(Available)(Ring *pRing)
{
  return pRing->_dwAvailable;
}
#endif // ]

////////////////
static BYTE *YASAPI_ID_RING(Buf2Ring)(Ring *pRing, BYTE *pbDst,
    const BYTE *pbSrc, DWORD dwSize)
{
	static double xSqrt2Inverse=0.0;

  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WORD wBytesPerSampleTarget=pTarget->wBitsPerSample>>3;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  WORD wBytesPerSampleSource=pSource->wBitsPerSample>>3;
  BYTE *pbDstMin=pbDst;
  BYTE *pbDstMax=pbDst+dwSize;
  WORD wChannel;

  int32_t i;
  BYTE *iwp=(BYTE *)&i+((sizeof i)-wBytesPerSampleSource);
  const BYTE *irp=(BYTE *)&i+((sizeof i)-wBytesPerSampleTarget);

  if (pTarget->nChannels==2&&pSource->nChannels==1) {
    switch (pEngine->options.common.eMono2Stereo) {
    case eMono2StereoHalf:
      while (pbDst<pbDstMax) {
        i=0;
        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        i>>=1;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }

      break;
    case eMono2StereoSqrt2Inverse:
      if (0.0==xSqrt2Inverse)
        xSqrt2Inverse=1.0/sqrt(2.0);

      while (pbDst<pbDstMax) {
        i=0;
        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        i=xSqrt2Inverse*i+0.5;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }

      break;
    default:
      if (wBytesPerSampleTarget==wBytesPerSampleSource) {
        while (pbDst<pbDstMax) {
          memcpy(pbDst,pbSrc,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
          memcpy(pbDst,pbSrc,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;

          pbSrc+=wBytesPerSampleTarget;
        }
      }
      else {
        while (pbDst<pbDstMax) {
          i=0;
          memcpy(iwp,pbSrc,wBytesPerSampleSource);
          pbSrc+=wBytesPerSampleSource;

          memcpy(pbDst,irp,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
          memcpy(pbDst,irp,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
        }
      }

      break;
    }
  }
  else if (pTarget->nChannels!=pSource->nChannels) {
    DMESSAGE("Multi channel mixing not supported");
    goto exit;
  }
  else {
    while (pbDst<pbDstMax) {
      for (wChannel=0u;wChannel<pTarget->nChannels;++wChannel) {
        i=0;
        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }
    }
  }
exit:
  return pbDstMin;
}

////////////////
IRingVtbl YASAPI_ID_IRING_VTBL()={
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
  .source.ConvertSize=YASAPI_ID_RING_SOURCE(ConvertSize),
#endif // ]
  .source.Size=YASAPI_ID_RING_SOURCE(Size),
  .source.Written=YASAPI_ID_RING_SOURCE(Written),
  .source.Available=YASAPI_ID_RING_SOURCE(Available),
  ////////////
  .target.ConvertSize=YASAPI_ID_RING_TARGET(ConvertSize),
  .target.Size=YASAPI_ID_RING_TARGET(Size),
  .target.Written=YASAPI_ID_RING_TARGET(Written),
#if defined (YASAPI_TARGET_AVAILABLE) // [
  .target.Available=YASAPI_ID_RING_TARGET(Available),
#endif // ]
  ////////////
  .Buf2Ring=YASAPI_ID_RING(Buf2Ring),
};
