/*
 * yasapi_thread.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (ERROR_INVALID_TASK_NAME) // [
#define ERROR_INVALID_TASK_NAME 1550L
#endif // ]

#if ! defined (ERROR_INVALID_TASK_INDEX) // [
#define ERROR_INVALID_TASK_INDEX 1551L
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef enum _EThreadLimits EThreadLimits;
typedef struct _Thread Thread;

///////////////////////////////////////////////////////////////////////////////
enum _EThreadLimits {
  eThreadCount=3,
};

struct _Thread {
  Engine *pEngine;
  HANDLE aHandles[eThreadCount];
  Command command;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  HANDLE hTask;
  DWORD dwTaskIndex;
#endif // ]
};

///////////////////////////////////////////////////////////////////////////////
static void ThreadCreate(Thread *pThread, Engine *pEngine)
{
  pThread->pEngine=pEngine;
  pThread->aHandles[0]=pEngine->thread.hProducer;
  pThread->aHandles[1]=pEngine->timer.hEvent;
  pThread->aHandles[2]=pEngine->timer.hTimer;
  pThread->command.eCommand=eCommandNull;
  pThread->command.eArgument=eEngineNull;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  pThread->hTask=NULL;
  pThread->dwTaskIndex=0ul;
#endif // ]

  EnterCriticalSection(&pEngine->cs); // {
  ++pEngine->eState;
  DASSERT(eEngineThreadRunning==pEngine->eState,exit);
  DWRITELN(EngineState(pEngine));
  SetEvent(pEngine->thread.hConsumer);
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return;
}

static void ThreadDestroy(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;

  DWRITELN(EngineState(pEngine));
  LeaveCriticalSection(&pEngine->cs); // }
  ExitThread(0ul);
}

static DWORD ThreadWait(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;
  DWORD dwWait;

  LeaveCriticalSection(&pEngine->cs); // {

  dwWait=WaitForMultipleObjects(
    eThreadCount,       // _In_       DWORD  nCount,
    pThread->aHandles,  // _In_ const HANDLE *lpHandles,
    FALSE,              // _In_       BOOL   bWaitAll,
    INFINITE            // _In_       DWORD  dwMilliseconds
  );

  EnterCriticalSection(&pEngine->cs); // }

  return dwWait;
}

static void ThreadKill(Thread *pThread, const EEngine eState)
{
  Engine *pEngine=pThread->pEngine;

  if (eEngineNull<eState) {
//DVWRITELN("going to kill %s -> %s <-----------------------",EngineState(pEngine),EngineStateEx(eState));
    EngineKill(pEngine,eState);
//DVWRITELN("killed %s / %s <-----------------------",EngineState(pEngine),EngineStateEx(eState));
  }

#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  if (pEngine->eState<eEngineAudioClient&&pThread->hTask) {
    AvRevertMmThreadCharacteristics(pThread->hTask);
    pThread->hTask=NULL;
    pThread->dwTaskIndex=0ul;
  }
#endif // ]

  DWRITELN(EngineState(pEngine));
  pEngine->thread.pCommand=NULL;
  SetEvent(pEngine->thread.hConsumer);
}

static int ThreadCommand(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
  ECommand eCommand=pThread->command.eCommand;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  DWORD dwLastError;
#endif // ]

  switch (eCommand) {
  case eCommandStart:
    if (eEngineStarted<=pEngine->eState)
      goto estart;

#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    if (pEngine->options.device.bProAudio&&!pThread->hTask) {
      pThread->hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",
          &pThread->dwTaskIndex);

      if (!pThread->hTask) {
        dwLastError=GetLastError();

        switch (dwLastError) {
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_INDEX);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_NAME);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_PRIVILEGE_NOT_HELD);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_ACCESS_DENIED);
        default:
          DVMESSAGE("AvSetMmThreadCharacteristics failed with"
              " unknow error %lu",dwLastError);
          break;
        }
      }
    }
#endif // ]

#if 0 // [
    if (eEngineAudioClient<=pEngine->eState) {
DVWRITELN("%s =========================================",EngineState(pEngine));
      if (EnginePlayingStart(pEngine)<0) {
        DMESSAGE("start playing");
        goto exit;
      }
DVWRITELN("%s =========================================",EngineState(pEngine));
    }
    else
#endif // ]
    EnginePlay(pEngine);
  estart:
    DWRITELN(EngineState(pEngine));
    pEngine->thread.pCommand=NULL;
    SetEvent(pEngine->thread.hConsumer);
    break;
  default:
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
#if 0 // [
exit:
#endif // ]
  return code;
}

static int ThreadPlay(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
  ECommand *peCommand=&pThread->command.eCommand;
  EEngine *peArgument=&pThread->command.eArgument;
  Ring *pRing=&pEngine->ring;

  switch (pEngine->eState) {
  case eEngineEot:
  case eEnginePauseConnected:
  case eEnginePauseDisconnected:
    break;
  case eEngineDraining:
    if (0ul==pRing->lpVtbl->target.Written(pRing)) {
      ++pEngine->eState;
      DWRITELN(EngineState(pEngine));
      DASSERT(eEngineEot==pEngine->eState,exit);
#if 0 // [
      break;
#endif // ]
    }

    // Intentional fall-through.
  default:
    if (eEngineStarted<=pEngine->eState&&EnginePlay(pEngine)<0) {
      DMESSAGE("playing");
      goto exit;
    }

    break;
  }

  if (*peCommand) {
    // Apply deferred command.
    switch (*peCommand) {
    case eCommandFlush:
      EngineFlush(pEngine);
      break;
    case eCommandKill:
      ThreadKill(pThread,*peArgument);
      break;
    default:
      break;
    }
  }

  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
DWORD WINAPI EngineThreadProc(LPVOID lpParameter)
{
  Engine *pEngine=lpParameter;
  Thread thread;
  ECommand *peCommand=&thread.command.eCommand;
  EEngine *peState=&thread.command.eArgument;
  DWORD dwWait;

  ThreadCreate(&thread,lpParameter); // {

  for (;;) {
    dwWait=ThreadWait(&thread);

    switch (dwWait) {
    case WAIT_OBJECT_0+0ul:
      DASSERT(pEngine->thread.pCommand,exit);
      thread.command=*pEngine->thread.pCommand;

      if (eCommandExit==*peCommand) {
        pEngine->thread.pCommand=NULL;
        SetEvent(pEngine->thread.hConsumer);
        goto exit;
      }
      else if (eCommandKill==*peCommand&&eEngineThreadRunning==*peState) {
        // The command is issued when not playing!!!
        ThreadKill(&thread,eEngineThreadRunning);
      }
      else if (ThreadCommand(&thread)<0)
        goto exit;

      pEngine->thread.pCommand=NULL;
      SetEvent(pEngine->thread.hConsumer);
      break;
    case WAIT_OBJECT_0+1ul:
    case WAIT_OBJECT_0+2ul:
      if (ThreadPlay(&thread)<0) {
        DMESSAGE("playing");
        goto exit;
      }

      break;
    default:
      DVMESSAGE("wait error: %lu",dwWait);
      goto exit;
    }
  }
exit:
  ThreadDestroy(&thread); // }
  return 0ul;
}
