/*
 * yasapi_wasapi.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

#if defined (YASAPI_DEVICE_MARKUP) // [
static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx,
    int bMarkup);
#else // ] [
static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx);
#endif // ]

int WasapiCreate(Wasapi *pWasapi, Player *pPlayer)
{
  HRESULT hr;
  UINT nDevice;

  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,      // _In_  REFCLSID  rclsid,
    NULL,                           // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                     // _In_  DWORD     dwClsContext,
    &IID_IMMDeviceEnumerator,       // _In_  REFIID    riid,
    (void**)&pWasapi->pEnumerator   // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) switch (hr) {
  case REGDB_E_CLASSNOTREG:
    DMESSAGE("CoCreateInstance() failed with REGDB_E_CLASSNOTREG");
    goto eenumerator;
  case CLASS_E_NOAGGREGATION:
    DMESSAGE("CoCreateInstance() failed with CLASS_E_NOAGGREGATION");
    goto eenumerator;
  case E_NOINTERFACE:
    DMESSAGE("CoCreateInstance() failed with E_NOINTERFACE");
    goto eenumerator;
  case E_POINTER:
    DMESSAGE("CoCreateInstance() failed with E_POINTER");
    goto eenumerator;
  case 0x800401f0:
    DMESSAGE("CoCreateInstance() failed with COM not initialized");
    goto eenumerator;
  default:
    DVMESSAGE("CoCreateInstance() failed with 0x%p/%lu",hr,hr);
    goto eenumerator;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pWasapi->pEnumerator->lpVtbl->EnumAudioEndpoints(pWasapi->pEnumerator,
    eRender,                        // [in]  EDataFlow           dataFlow,
    DEVICE_STATE_ACTIVE,            // [in]  DWORD               dwStateMask,
    &pWasapi->pCollection           // [out] IMMDeviceCollection **ppDevices
  );

  if (FAILED(hr)) switch (hr) {
  case E_POINTER:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_POINTER");
    goto ecollection;
  case E_INVALIDARG:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_INVALIDARG");
    goto ecollection;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with"
        " E_OUTOFMEMORY");
    goto ecollection;
  default:
    DVMESSAGE("IMMDeviceEnumerator::EnumAudioEndpoint() failed with %p",hr);
    goto ecollection;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pWasapi->pCollection->lpVtbl->GetCount(pWasapi->pCollection,
    &pWasapi->cDevices              // [out] UINT *pcDevices
  );

  if (FAILED(hr)) switch (hr) {
  case E_POINTER:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_POINTER");
    goto ecount;
  case E_INVALIDARG:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_INVALIDARG");
    goto ecount;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDeviceCollection::GetCount failed with E_OUTOFMEMORY");
    goto ecount;
  default:
    DVMESSAGE("IMMDeviceCollection::GetCount() failed with %p",hr);
    goto ecount;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (OptionsCreate(&pWasapi->options,pPlayer,pWasapi->cDevices)<0) {
    DMESSAGE("Creating options.");
    goto eoptions;
  }

  /////////////////////////////////////////////////////////////////////////////
  ++pWasapi->cDevices;
  pWasapi->pDevices=malloc(pWasapi->cDevices*sizeof *pWasapi->pDevices);

  if (!pWasapi->pDevices) {
    DMESSAGE("allocating WasapiDevice array");
    goto emalloc;
  }

  for (nDevice=0;nDevice<pWasapi->cDevices;++nDevice) {
    if (WasapiDeviceCreate(pWasapi->pDevices+nDevice,pWasapi,nDevice-1)<0) {
      DMESSAGE("creating WasapiDevice");
      goto edevice;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (WasapiSelectDevice(pWasapi,NULL)<0) {
    DMESSAGE("selecting device");
    goto eselect;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//error:
eselect:
edevice:
  while (0<nDevice) {
    --nDevice;
    WasapiDeviceDestroy(pWasapi->pDevices+nDevice,&pWasapi->options);
  }

  pWasapi->cDevices=0u;
  free(pWasapi->pDevices);
emalloc:
  OptionsDestroy(&pWasapi->options);
eoptions:
ecount:
  pWasapi->pCollection->lpVtbl->Release(pWasapi->pCollection);
ecollection:
  pWasapi->pEnumerator->lpVtbl->Release(pWasapi->pEnumerator);
eenumerator:
  return -1;
}

void WasapiDestroy(Wasapi *pWasapi, Player *pPlayer)
{
  UINT nDevice=pWasapi->cDevices;

  while (0u<nDevice) {
    --nDevice;
    WasapiDeviceDestroy(pWasapi->pDevices+nDevice,&pWasapi->options);
  }

  free(pWasapi->pDevices);
  OptionsDestroy(&pWasapi->options);
  pWasapi->pCollection->lpVtbl->Release(pWasapi->pCollection);
  pWasapi->pEnumerator->lpVtbl->Release(pWasapi->pEnumerator);
}

int WasapiSelectDevice(Wasapi *pWasapi, WasapiDevice *pDevice)
{
  wchar_t *lhs=pWasapi->options.common.current.id;
  const wchar_t *rhs;
  UINT nDevice;

  if (pDevice) {
    //pWasapi->selected.pDevice=pDevice;
    pWasapi->selected.pDeviceConfig=pDevice;
    rhs=pDevice->lpVtbl->getid(pDevice);
    wcscpy(lhs,rhs);
  }
  else {
    if (0u==pWasapi->cDevices) {
      DMESSAGE("no devices available");
      pWasapi->selected.pDevice=NULL;
      pWasapi->selected.pDeviceConfig=NULL;
      goto error;;
    }

    for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
      pDevice=pWasapi->pDevices+nDevice;
      rhs=pDevice->lpVtbl->getid(pDevice);

      if (!wcscmp(lhs,rhs)) {
        pWasapi->selected.pDevice=pDevice;
        pWasapi->selected.pDeviceConfig=pDevice;
        goto success;
      }
    }

    pDevice=pWasapi->pDevices;
    pWasapi->selected.pDevice=pDevice;
    pWasapi->selected.pDeviceConfig=pDevice;
    rhs=pDevice->lpVtbl->getid(pDevice);
    wcscpy(lhs,rhs);
  }
success:
  DVWRITELNW(L"selected=\"%s\"",lhs);
  return 0;
error:
  return -1;
}

void WasapiRestore(Wasapi *pWasapi)
{
  UINT nDevice;
  WasapiDevice *pDevice;

  pWasapi->options.common.config=pWasapi->options.common.current;

  for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
    pDevice=pWasapi->pDevices+nDevice;
    pDevice->pMember->config=pDevice->pMember->current;
  }

  pWasapi->selected.pDeviceConfig=pWasapi->selected.pDevice;
}

void WasapiSync(Wasapi *pWasapi)
{
  UINT nDevice;
  WasapiDevice *pDevice;

  pWasapi->options.common.current=pWasapi->options.common.config;

  for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
    pDevice=pWasapi->pDevices+nDevice;
    pDevice->pMember->current=pDevice->pMember->config;
  }

  pWasapi->selected.pDevice=pWasapi->selected.pDeviceConfig;
}

void WasapiDefault(Wasapi *pWasapi)
{
  UINT nDevice;
  WasapiDevice *pDevice;

  pWasapi->options.common.config=gcOptionsCommon;

  for (nDevice=0u;nDevice<pWasapi->cDevices;++nDevice) {
    pDevice=pWasapi->pDevices+nDevice;
    pDevice->pMember->config=gcOptionsDevice;
  }

  pWasapi->selected.pDeviceConfig=pWasapi->pDevices;
}

///////////////////////////////////////////////////////////////////////////////
static const wchar_t *WasaspiDeviceGetId(WasapiDevice *pDevice)
{
  return pDevice->id;
}

static IWasapiDeviceVtbl gcIWasapiDeviceVtbl={
#if defined (YASAPI_DEBUG) // [
  .pId="WasapiDeviceVtbl",
#endif // ]
  .getid=WasaspiDeviceGetId,
};

///////////////////////////////////////////////////////////////////////////////
static const wchar_t *WasaspiDeviceDefaultGetId(WasapiDevice *pDevice)
{
  return L"";
}

static IWasapiDeviceVtbl gcIWasapiDeviceVtblDefault={
#if defined (YASAPI_DEBUG) // [
  .pId="WasapiDeviceVtblDefault",
#endif // ]
  .getid=WasaspiDeviceDefaultGetId,
};

///////////////////////////////////////////////////////////////////////////////
int WasapiDeviceCreate(WasapiDevice *pDevice, Wasapi *pWasapi, int nDevice)
{
  HRESULT hr;

  if (nDevice<0) {
    ///////////////////////////////////////////////////////////////////////////
    pDevice->lpVtbl=&gcIWasapiDeviceVtblDefault;

    hr=pWasapi->pEnumerator->lpVtbl->GetDefaultAudioEndpoint(
      pWasapi->pEnumerator,
      eRender,                      // [in]  EDataFlow dataFlow,
      eConsole,                     // [in]  ERole     role,
      &pDevice->pDevice             // [out] IMMDevice **ppDevice
    );

    if (FAILED(hr))
#if defined (_MSC_VER) // [
      if (E_NOTFOUND==hr) {
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_NOTFOUND");
        goto edevice;
      }
      else
#endif
      switch (hr) {
      case E_POINTER:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_POINTER");
        goto edevice;
      case E_INVALIDARG:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_INVALIDARG");
        goto edevice;
#if ! defined (_MSC_VER) // [
      case E_NOTFOUND:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_NOTFOUND");
        goto edevice;
#endif // ]
      case E_OUTOFMEMORY:
        DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with E_OUTOFMEMORY");
        goto edevice;
      default:
        DVMESSAGE("IIMMDeviceEnumerator::GetDefaultAudioEndpoint() failed"
            " with %p",hr);
        goto edevice;
      }
  }
  else {
    ///////////////////////////////////////////////////////////////////////////
    pDevice->lpVtbl=&gcIWasapiDeviceVtbl;

    ///////////////////////////////////////////////////////////////////////////
    hr=pWasapi->pCollection->lpVtbl->Item(pWasapi->pCollection,
      nDevice,                      // [in]  UINT      nDevice,
      &pDevice->pDevice             // [out] IMMDevice **ppDevice
    );

    if (FAILED(hr)) switch (hr) {
    case E_POINTER:
      DMESSAGE("IMMDeviceCollection::Item() failed with E_POINTER");
      goto edevice;
    case E_INVALIDARG:
      DMESSAGE("IMMDeviceCollection::Item() failed with E_INVALIDARG");
      goto edevice;
    default:
      DVMESSAGE("IMMDeviceCollection::Item() failed with 0x%p",hr);
      goto edevice;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pDevice->pDevice->lpVtbl->GetId(pDevice->pDevice,
    &pDevice->id                    // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) switch (hr) {
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::GetId() failed with E_OUTOFMEMORY");
    goto eid;
  case E_POINTER:
    DMESSAGE("IMMDevice::GetId() failed with E_POINTER");
    goto eid;
  default:
    DVMESSAGE("IMMDevice::GetId() failed with 0x%p",hr);
    goto eid;
  }

  /////////////////////////////////////////////////////////////////////////////
  pDevice->pMember=OptionsSearch(&pWasapi->options,pDevice->id);

  if (!pDevice->pMember) {
    DMESSAGE("options not found");
    goto emember;
  }

  OptionsMemberDeviceRef(pDevice->pMember);

  /////////////////////////////////////////////////////////////////////////////
  hr=pDevice->pDevice->lpVtbl->OpenPropertyStore(pDevice->pDevice,
    STGM_READ,                      // [in]  DWORD          stgmAccess,
    &pDevice->pStore                // [out] IPropertyStore **ppProperties
  );

  if (FAILED(hr)) switch (hr) {
  case E_INVALIDARG:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_INVALIDARG");
    goto estore;
  case E_POINTER:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_POINTER");
    goto estore;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::OpenPropertyStore() failed with E_OUTOFMEMORY");
    goto estore;
  default:
    DVMESSAGE("IMMDevice::OpenPropertyStore() failed with %p",hr);
    goto estore;
  }

  /////////////////////////////////////////////////////////////////////////////
  PropVariantInit(&pDevice->vFriendlyName);

  hr=pDevice->pStore->lpVtbl->GetValue(pDevice->pStore,
    &PKEY_Device_FriendlyName,      // [in]  REFPROPERTYKEY key,
    &pDevice->vFriendlyName         // [out] PROPVARIANT    *pv
  );

  if (FAILED(hr)) {
    DVMESSAGE("IPropertyStore::GetValue() failed with %p",hr);
    goto efriendly;
  }

  /////////////////////////////////////////////////////////////////////////////
  pDevice->pFriendlyNameOriginal=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
      NULL,NULL,0);

  if (!pDevice->pFriendlyNameOriginal) {
    DMESSAGE("converting FriendlyName to UTF-8");
    goto eoriginal;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (nDevice<0) {
#if defined (YASAPI_DEVICE_MARKUP) // [
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        "Default Device [","]",0);
    pDevice->pFriendlyNameMarkup=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        "Default Device [","]",1);
#else // ] [
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        "Default Device [","]");
#endif // ]
  }
  else {
#if defined (YASAPI_DEVICE_MARKUP) // [
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        NULL,NULL,0);
    pDevice->pFriendlyNameMarkup=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        NULL,NULL,1);
#else // ] [
    pDevice->pFriendlyName=_wcs2strdup(pDevice->vFriendlyName.pwszVal,
        NULL,NULL);
#endif // ]
  }

  if (!pDevice->pFriendlyName) {
    DMESSAGE("converting FriendlyName to UTF-8");
    goto eclone;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (1u==pDevice->pMember->uRef) {
    if (FormatMatrixCreate(&pDevice->pMember->matrix,pDevice->pDevice)<0) {
      DMESSAGE("creating matrix");
      goto ematrix;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//error:
  if (1u==pDevice->pMember->uRef)
    FormatMatrixDestroy(&pDevice->pMember->matrix);
ematrix:
  free(pDevice->pFriendlyName);
eclone:
  free(pDevice->pFriendlyNameOriginal);
eoriginal:
  PropVariantClear(&pDevice->vFriendlyName);
efriendly:
  pDevice->pStore->lpVtbl->Release(pDevice->pStore);
estore:
  OptionsMemberDeviceUnref(pDevice->pMember,&pWasapi->options);
emember:
  CoTaskMemFree(pDevice->id);
eid:
  pDevice->pDevice->lpVtbl->Release(pDevice->pDevice);
edevice:
  return -1;
}

void WasapiDeviceDestroy(WasapiDevice *pDevice, Options *pOptions)
{
  if (0u==--pDevice->pMember->uRef)
    FormatMatrixDestroy(&pDevice->pMember->matrix);

  free(pDevice->pFriendlyName);
  free(pDevice->pFriendlyNameOriginal);
  PropVariantClear(&pDevice->vFriendlyName);
  pDevice->pStore->lpVtbl->Release(pDevice->pStore);
  OptionsMemberDeviceUnref(pDevice->pMember,pOptions);
  CoTaskMemFree(pDevice->id);
  pDevice->pDevice->lpVtbl->Release(pDevice->pDevice);
}

///////////////////////////////////////////////////////////////////////////////
int FormatMatrixCreate(FormatMatrix *pMatrix, IMMDevice *pDevice)
{
  WAVEFORMATEX *pClosestMatch=NULL;
  HRESULT hr;
  IAudioClient *pClient;
  int row,col;
  WAVEFORMATEXTENSIBLE wfx;

  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,  // [in]  REFIID      iid,
    CLSCTX_ALL,         // [in]  DWORD       dwClsCtx,
    NULL,               // [in]  PROPVARIANT *pActivationParams,
    (void **)&pClient   // [out] void        **ppInterface
  );
  
  if (FAILED(hr)) switch (hr) {
  case E_NOINTERFACE:
    DMESSAGE("IMMDevice::Activate() failed with E_NOINTERFACE");
    goto eactivate;
  case E_POINTER:
    DMESSAGE("IMMDevice::Activate() failed with E_POINTER");
    goto eactivate;
  case E_INVALIDARG:
    DMESSAGE("IMMDevice::Activate() failed with E_INVALIDARG");
    goto eactivate;
  case E_OUTOFMEMORY:
    DMESSAGE("IMMDevice::Activate() failed with E_OUTOFMEMORY");
    goto eactivate;
  case AUDCLNT_E_DEVICE_INVALIDATED:
    DMESSAGE("IMMDevice::Activate() failed with AUDCLNT_E_DEVICE_INVALIDATED");
    goto eactivate;
  default:
    DVMESSAGE("IMMDevice::Activate() failed with %p",hr);
    goto eactivate;
  }

  wfx.Format.wFormatTag=WAVE_FORMAT_EXTENSIBLE;
  wfx.Format.nChannels=2;
  wfx.Format.cbSize=(sizeof wfx)-sizeof (wfx.Format);
  ////
  wfx.dwChannelMask=SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT;
  wfx.SubFormat=KSDATAFORMAT_SUBTYPE_PCM;

  for (row=0;row<eFrequenciesNum;++row) {
    wfx.Format.nSamplesPerSec=gaFrequencies[row].freq;;

    for (col=0;col<eBitDepthsNum;++col) {
      wfx.Format.wBitsPerSample=gaBitDepths[col].bits;
      wfx.Format.nBlockAlign=wfx.Format.nChannels*wfx.Format.wBitsPerSample;
      wfx.Format.nAvgBytesPerSec
          =wfx.Format.nSamplesPerSec*wfx.Format.nBlockAlign;
      ////
      wfx.Samples.wValidBitsPerSample=wfx.Format.wBitsPerSample;

      hr=pClient->lpVtbl->IsFormatSupported(pClient,
                        // [in]        AUDCLNT_SHAREMODE ShareMode,
        AUDCLNT_SHAREMODE_EXCLUSIVE,
        &wfx.Format,    // [in]  const WAVEFORMATEX      *pFormat,
        &pClosestMatch  // [out]       WAVEFORMATEX      **ppClosestMatch
      );

      if (pClosestMatch) {
        CoTaskMemFree(pClosestMatch);
        pClosestMatch=NULL;
      }

      pMatrix->bSupported[row][col]=0;

      switch (hr) {
      case S_OK:
        pMatrix->bSupported[row][col]=1;
        break;
      case S_FALSE:
      case AUDCLNT_E_UNSUPPORTED_FORMAT:
        break;
      case E_POINTER:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " E_POINTER");
        break;
      case E_INVALIDARG:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " E_INVALIDARG");
        break;
      case AUDCLNT_E_DEVICE_INVALIDATED:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " AUDCLNT_E_DEVICE_INVALIDATED");
        break;
      case AUDCLNT_E_SERVICE_NOT_RUNNING:
        DMESSAGE("IAudioClient::IsFormatSupported() failed with"
            " AUDCLNT_E_SERVICE_NOT_RUNNING");
        break;
      default:
        DVMESSAGE("IAudioClient::IsFormatSupported() failed with %p",hr);
        break;
      }
    }
  }

  pClient->lpVtbl->Release(pClient);

  return 0;
eactivate:
  return -1;
}

void FormatMatrixDestroy(FormatMatrix *pMatrix)
{
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEVICE_MARKUP) // [
static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx,
    int bMarkup)
#else // ] [
static char *_wcs2strdup(wchar_t *wcs, const char *pfx, const char *sfx)
#endif // ]
{
#if defined (YASAPI_DEVICE_MARKUP) // [
  const char BL[]="<b>";
  const char BR[]="</b>";

  enum {
    eLenBL = (sizeof BL) - 1,
    eLenBR = (sizeof BR) - 1,
    eLenB=eLenBL+eLenBR,
  };
#endif // ]

  int nLen1=pfx?strlen(pfx):0;
  int nSize2;
  int nLen3=sfx?strlen(sfx):0;
  int nSize;
  char *str,*p;

  if (!wcs)
    goto ewcs;

  nSize2=WideCharToMultiByte(
    CP_UTF8,        // _In_      UINT    CodePage,
    0ul,            // _In_      DWORD   dwFlags,
    wcs,            // _In_      LPCWSTR lpWideCharStr,
    -1,             // _In_      int     cchWideChar,
    NULL,           // _Out_opt_ LPSTR   lpMultiByteStr,
    0,              // _In_      int     cbMultiByte,
    NULL,           // _In_opt_  LPCSTR  lpDefaultChar,
    NULL            // _Out_opt_ LPBOOL  lpUsedDefaultChar
  );

  nSize=nLen1+nSize2+nLen3;
#if defined (YASAPI_DEVICE_MARKUP) // [
  if (bMarkup)
    nSize+=eLenB;
#endif // ]

  if (!(str=malloc(nSize)))
    goto emalloc;

  p=str;

#if defined (YASAPI_DEVICE_MARKUP) // [
  if (bMarkup) {
    strcpy(p,BL);
    p+=eLenBL;
  }
#endif // ]

  if (0<nLen1) {
    strcpy(p,pfx);
    p+=nLen1;
  }

  WideCharToMultiByte(
    CP_UTF8,        // _In_      UINT    CodePage,
    0ul,            // _In_      DWORD   dwFlags,
    wcs,            // _In_      LPCWSTR lpWideCharStr,
    -1,             // _In_      int     cchWideChar,
    p,              // _Out_opt_ LPSTR   lpMultiByteStr,
    nSize2,         // _In_      int     cbMultiByte,
    NULL,           // _In_opt_  LPCSTR  lpDefaultChar,
    NULL            // _Out_opt_ LPBOOL  lpUsedDefaultChar
  );

  p+=nSize2-1;

  if (0<nLen3) {
    strcpy(p,sfx);
    p+=nLen3;
  }

#if defined (YASAPI_DEVICE_MARKUP) // [
  if (bMarkup) {
    strcpy(p,BR);
    p+=eLenBR;
  }
#endif // ]

  return str;
emalloc:
ewcs:
  return NULL;
}
